# Copyright 2024 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from types import NoneType as _NoneType
from typing import Any as _Any, Collection as _Collection

from simplevent import RefEvent as _RefEvent
from waapi import EventHandler as _EventHandler, WaapiClient as _WaapiClient

from pywwise.aliases import ListOrTuple
from pywwise.decorators import callback
from pywwise.enums import (EAttenuationCurveShape, EAttenuationCurveType, EAttenuationCurveUsage, EListMode,
                           ENameConflictStrategy, EObjectType, EPropertyPasteMode, EReturnOptions, ERtpcMode)
from pywwise.primitives import GUID, Name, ProjectPath
from pywwise.statics import EnumStatics
from pywwise.structs import (AttenuationCurve, GraphPoint2D, PropertyInfo, SetOperation, Vector2, WwiseObjectInfo,
                             WwiseObjectWatch)
from pywwise.waql import WaqlQuery


class Object:
    """ak.wwise.core.object"""
    
    def __init__(self, client: _WaapiClient, watch_list: tuple[WwiseObjectWatch, ...] = ()):
        """
        Constructor.
        :param client: The WAAPI client to use.
        :param watch_list: A tuple of `WwiseObjectWatch` instances. This will be used to set up the
                           `ak.wwise.core.object.property_changed` event.
        """
        self._client = client
        
        return_options = {"return": [EReturnOptions.GUID, EReturnOptions.NAME,
                                     EReturnOptions.TYPE, EReturnOptions.PATH]}
        
        self.attenuation_curve_changed = _RefEvent(WwiseObjectInfo, EAttenuationCurveType)
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_attenuationcurvechanged.html
        \nSent when an attenuation curve is changed.
        \n**Event Data**:
        \n- A WwiseObjectInfo instance representing the Attenuation object that owns the curve.
        \n- The type of the curve that changed.
        """
        
        self._attenuation_curve_changed = self._client.subscribe(
            "ak.wwise.core.object.attenuationCurveChanged",
            self._on_attenuation_curve_changed, return_options)
        
        self.attenuation_curve_link_changed = _RefEvent(WwiseObjectInfo, EAttenuationCurveType)
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_attenuationcurvelinkchanged.html
        \nSent when an attenuation curve's link/unlink is changed. NOTE: this event often multi-triggers.
        \n**Event Data**:
        \n- A WwiseObjectInfo instance representing the Attenuation object that owns the curve.
        \n- The type of the curve that had its link changed.
        """
        
        self._attenuation_curve_link_changed = self._client.subscribe(
            "ak.wwise.core.object.attenuationCurveLinkChanged",
            self._on_attenuation_curve_link_changed, return_options)
        
        self.child_added = _RefEvent(WwiseObjectInfo, WwiseObjectInfo)
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_childadded.html
        \nSent when an object is added as a child to another object.
        \n*Event Data**:
        \n- A WwiseObjectInfo instance representing the new child object.
        \n- A WwiseObjectInfo instance representing the parent of the new child object.
        """
        
        self._child_added = self._client.subscribe("ak.wwise.core.object.childAdded",
                                                   self._on_child_added, return_options)
        
        self.child_removed = _RefEvent(WwiseObjectInfo, WwiseObjectInfo)
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_childadded.html
        \nSent when an object is removed from the children of another object.
        \n*Event Data**:
        \n- A WwiseObjectInfo instance representing the child object that got removed.
        \n- A WwiseObjectInfo instance representing the former parent of the child object.
        """
        
        self._child_removed = self._client.subscribe("ak.wwise.core.object.childRemoved",
                                                     self._on_child_removed, return_options)
        
        self.created = _RefEvent(WwiseObjectInfo)
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_created.html
        \nSent when an object is created. The name and path are not available at the time of creation.
        \n**Event Data**:
        \n- A WwiseObject instance representing the newly created object.
        """
        
        self._created = self._client.subscribe("ak.wwise.core.object.created",
                                               self._on_created, return_options)
        
        self.curve_changed = _RefEvent(WwiseObjectInfo, WwiseObjectInfo)
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_curvechanged.html
        \nSent when one or many curves are changed.
        \n**Event Data**:
        \n- A WwiseObjectInfo instance representing the curve that changed.
        \n- A WwiseObjectInfo instance representing the owner of the curve that changed.
        """
        
        self._curve_changed = self._client.subscribe("ak.wwise.core.object.curveChanged",
                                                     self._on_curve_changed, return_options)
        
        self.name_changed = _RefEvent(WwiseObjectInfo, str)
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_namechanged.html
        \nSent when an object is renamed. Publishes the renamed object.
        \n**Event Data**:
        \n- A WwiseObjectInfo instance representing the object that was renamed. It will contain the new name.
        \n- A string containing the old name.
        """
        
        self._name_changed = self._client.subscribe("ak.wwise.core.object.nameChanged",
                                                    self._on_name_changed, return_options)
        
        self.notes_changed = _RefEvent(WwiseObjectInfo, str, str)
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_noteschanged.html
        \nSent when the object's notes are changed.
        \n**Event Data**:
        \n- A WwiseObjectInfo instance representing the object whose notes changed.
        \n- A string containing the old notes.
        \n- A string containing the new notes.
        """
        
        self._notes_changed = self._client.subscribe("ak.wwise.core.object.notesChanged",
                                                     self._on_notes_changed, return_options)
        
        self.post_deleted = _RefEvent(WwiseObjectInfo)
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_postdeleted.html
        \nSent following an object's deletion.
        \n**Event Data**:
        \n- A WwiseObjectInfo instance representing the object that was deleted.
        """
        
        self._post_deleted = self._client.subscribe("ak.wwise.core.object.postDeleted",
                                                    self._on_post_deleted, return_options)
        
        self.pre_deleted = _RefEvent(WwiseObjectInfo)
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_predeleted.html
        \nSent prior to an object's deletion.
        \n**Event Data**:
        \n- A WwiseObjectInfo instance representing the object that will be deleted.
        """
        
        self._pre_deleted = self._client.subscribe("ak.wwise.core.object.preDeleted",
                                                   self._on_pre_deleted, return_options)
        
        self.property_changed = _RefEvent(WwiseObjectInfo, Name, _Any, _Any, GUID)
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_propertychanged.html
        \nSent when the watched property of an object changes.
        \n**Event Data**:
        \n- A WwiseObjectInfo instance representing the watched object.
        \n- The Name of the changed property.
        \n- The old value of the property. Can be of any type, but likely `None`, `str`, `float`, or `bool`.
        \n- The new value of the property. Can be of any type, but likely `None`, `str`, `float`, or `bool`.
        \n- The GUID of the platform for which the change occurred.
        \n**Additional Notes**:
        \n- This event requires a `watch_list` (a `tuple` of `WwiseObjectWatch`). See `pywwise.new_connection`.
        \n- This event will only happen for the objects and properties included in the `watch_list`.
        """
        
        self._property_changed = list[_EventHandler]()
        for watch in watch_list:
            for prop in watch.properties:  # `property` is a built-in identifier, so using `prop` instead
                prop_options = dict(return_options)
                prop_options["object"] = watch.guid
                prop_options["property"] = prop
                self._property_changed.append(self._client.subscribe("ak.wwise.core.object.propertyChanged",
                                                                     self._on_property_changed, prop_options))
        
        self.reference_changed = _RefEvent(WwiseObjectInfo)
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_referencechanged.html
        \nSent when an object reference is changed.
        \n**Event Data**:
        \n- A WwiseObjectInfo instance representing the object that had a reference changed.
        \n- A WwiseObjectInfo instance representing the previous referenced object.
        \n- A WwiseObjectInfo instance representing the new referenced object.
        """
        
        self._reference_changed = self._client.subscribe("ak.wwise.core.object.referenceChanged",
                                                         self._on_reference_changed, return_options)
    
    @callback
    def _on_attenuation_curve_changed(self, event, **kwargs):
        """
        Callback function for the `attenuationCurveChanged` event.
        :param event: The event to broadcast.
        :param kwargs: The event data.
        """
        event(*self._on_attenuation_event(**kwargs))
    
    @callback
    def _on_attenuation_curve_link_changed(self, event, **kwargs):
        """
        Callback function for the `attenuationCurveLinkChanged` event.
        :param event: The event to broadcast.
        :param kwargs: The event data.
        """
        event(*self._on_attenuation_event(**kwargs))
    
    @staticmethod
    def _on_attenuation_event(**kwargs) -> tuple[WwiseObjectInfo, EAttenuationCurveType]:
        """
        Utility function for the `attenuationCurveChanged` and `attenuationCurveLinkChanged` events.
        :param kwargs: The event data.
        :return: The event data, processed.
        """
        info = WwiseObjectInfo.from_dict(kwargs["attenuation"])
        curve = EnumStatics.from_value(EAttenuationCurveType, kwargs["curveType"])
        return info, curve
    
    @callback
    def _on_child_added(self, event, **kwargs):
        """
        Callback function for the `childAdded` event.
        :param event: The event to broadcast.
        :param kwargs: The event data.
        """
        event(*self._on_child_event(**kwargs))
    
    @callback
    def _on_child_removed(self, event, **kwargs):
        """
        Callback function for the `childRemoved` event.
        :param event: The event to broadcast.
        :param kwargs: The event data.
        """
        event(*self._on_child_event(**kwargs))
    
    @staticmethod
    def _on_child_event(**kwargs) -> tuple[WwiseObjectInfo, WwiseObjectInfo]:
        """
        Utility function for the `childAdded` and `childRemoved` events.
        :param kwargs: The event data.
        :return: The event data, processed.
        """
        return WwiseObjectInfo.from_dict(kwargs["child"]), WwiseObjectInfo.from_dict(kwargs["parent"])
    
    @callback
    def _on_created(self, event, **kwargs):
        """
        Callback function for the `created` event.
        :param event: The event to broadcast.
        :param kwargs: The event data.
        """
        obj = kwargs["object"]
        event(WwiseObjectInfo(GUID(obj["id"]),
                              Name.get_null(),
                              EObjectType.from_type_name(obj["type"]),
                              ProjectPath.get_null()))
    
    @callback
    def _on_curve_changed(self, event, **kwargs):
        """
        Callback function for the `curveChanged` event.
        :param event: The event to broadcast.
        :param kwargs: The event data.
        """
        event(WwiseObjectInfo.from_dict(kwargs["curve"]), WwiseObjectInfo.from_dict(kwargs["owner"]))
    
    @callback
    def _on_name_changed(self, event, **kwargs):
        """
        Callback function for the `nameChanged` event.
        :param event: The event to broadcast.
        :param kwargs: The event data.
        """
        event(WwiseObjectInfo.from_dict(kwargs["object"]), kwargs["oldName"])
    
    @callback
    def _on_notes_changed(self, event, **kwargs):
        """
        Callback function for the `notesChanged` event.
        :param event: The event to broadcast.
        :param kwargs: The event data.
        """
        event(WwiseObjectInfo.from_dict(kwargs["object"]), kwargs["newNotes"], kwargs["oldNotes"])
    
    @callback
    def _on_post_deleted(self, event, **kwargs):
        """
        Callback function for the `postDeleted` event.
        :param event: The event to broadcast.
        :param kwargs: The event data.
        """
        event(WwiseObjectInfo.from_dict(kwargs["object"]))
    
    @callback
    def _on_pre_deleted(self, event, **kwargs):
        """
        Callback function for the `preDeleted` event.
        :param event: The event to broadcast.
        :param kwargs: The event data.
        """
        event(WwiseObjectInfo.from_dict(kwargs["object"]))
    
    @callback
    def _on_property_changed(self, event, **kwargs):
        """
        Callback function for the `propertyChanged` event.
        :param event: The event to broadcast.
        :param kwargs: The event data.
        """
        watch = WwiseObjectInfo.from_dict(kwargs["object"])  # this is the object where the change happened
        event(watch, Name(kwargs["property"]), kwargs["old"], kwargs["new"], GUID(kwargs["platform"]))
    
    @callback
    def _on_reference_changed(self, event, **kwargs):
        """
        Callback function for the `referenceChanged` event.
        :param event: The event to broadcast.
        :param kwargs: The event data.
        """
        obj = WwiseObjectInfo.from_dict(kwargs["object"])
        old = WwiseObjectInfo.from_dict(kwargs["old"])
        new = WwiseObjectInfo.from_dict(kwargs["new"])
        event(obj, old, new)
    
    def copy(self, obj: GUID | tuple[EObjectType, Name] | ProjectPath,
             parent: GUID | tuple[EObjectType, Name] | ProjectPath,
             name_conflict_strategy: ENameConflictStrategy = ENameConflictStrategy.FAIL,
             version_control_auto_checkout: bool = True,
             version_control_auto_add: bool = True) -> WwiseObjectInfo | None:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_copy.html \n
        Copies an object to the given parent. Note that if a Work Unit is copied, the operation cannot be undone and
        the project will be saved. This function calls `ak.wwise.core.object.get` to fetch the `path` and `type`
        attributes.
        :param obj: The GUID, typed name, or project path of the object to copy. Although using a name is supported,
                    only types with globally-unique names (e.g. `EObjectType.EVENT`) are supported.
        :param parent: The GUID, type and name, or project path of the new object's parent. Although using a name is
                       supported, only types with globally-unique names (e.g. `EObjectType.EVENT`) are supported.
        :param name_conflict_strategy: The strategy to use in case of a name conflict.
        :param version_control_auto_checkout: Determines if Wwise automatically performs a Checkout source control
                                              operation for affected work units and for the project. Only supported in
                                              Wwise 2023 or above.
        :param version_control_auto_add: Determines if Wwise automatically performs an Add source control operation for
                                         affected work units. Only supported in Wwise 2023 or above.
        :return: The new object as a WwiseObjectInfo, or None if the operation failed.
        """
        args = {"object": f"{obj[0].get_type_name()}:{obj[1]}" if isinstance(obj, tuple) else obj,
                "parent": f"{parent[0].get_type_name()}:{parent[1]}" if isinstance(parent, tuple) else parent,
                "onNameConflict": name_conflict_strategy,
                **({"autoCheckOutToSourceControl": False} if not version_control_auto_checkout else {}),
                **({"autoAddToSourceControl": False} if not version_control_auto_add else {})}
        options = {"return": EReturnOptions.get_defaults()}
        
        results = self._client.call("ak.wwise.core.object.copy", args, options=options)
        if results is None:
            return None
        
        obj_info = self.get(f"$ from object \"{results.get('id', GUID.get_null())}\" take 1")  # force single match
        return obj_info[0] if len(obj_info) > 0 else None  # WwiseObjectInfo has valid path and type attributes
    
    def create(self, name: Name | str, etype: EObjectType, parent: GUID | tuple[EObjectType, Name] | ProjectPath,
               name_conflict_strategy: ENameConflictStrategy = ENameConflictStrategy.FAIL, notes: str = "",
               version_control_auto_add: bool = True, platform: Name | GUID = None) -> WwiseObjectInfo | None:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_create.html \n
        Creates an object of type 'type', as a child of 'parent'. Refer to Importing Audio Files and Creating
        Structures for more information about creating objects. Also refer to `ak.wwise.core.audio.import_files` to
        import audio files to Wwise. To create Effect or Source plug-ins, use `ak.wwise.core.object.set`, and refer to
        Wwise Objects Reference for the classId.
        :param name: The name of the new Wwise object.
        :param etype: The type of the new Wwise object.
        :param parent: The GUID, typed name, or project path of the new object's parent.
        :param name_conflict_strategy: The strategy to adopt in case of name conflicts.
        :param notes: The notes to append to the new Wwise object.
        :param version_control_auto_add: Determines if Wwise automatically performs a Checkout source control operation
                                         for affected work units and for the project. Only supported in Wwise 2023 or
                                         above.
        :param platform: Specify what platform to create the object for. Usually not necessary.
        :return: The new object as a WwiseObjectInfo, or None if the operation failed.
        """
        args = {"name": name, "type": etype.get_type_name(),
                "parent": f"{parent[0].get_type_name()}:{parent[1]}" if isinstance(parent, tuple) else parent,
                "onNameConflict": name_conflict_strategy.value,
                "notes": notes,
                **({"autoAddToSourceControl": False} if not version_control_auto_add else {})}
        if platform is not None:
            args["platform"] = platform
        
        results = self._client.call("ak.wwise.core.object.create", args)  # missing path, at this point
        if results is None:
            return None
        
        new_obj = self.get(f"$ from object \"{results.get('id', GUID.get_null())}\" take 1")
        return new_obj[0] if len(new_obj) > 0 else None
    
    def delete(self, obj: GUID | tuple[EObjectType, Name] | ProjectPath,
               version_control_auto_checkout: bool = True) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_delete.html \n
        Deletes the specified object. Note that if a Work Unit is deleted, the operation cannot be undone
        and the project will be saved.
        :param obj: The GUID, typed name, or project path of the object to delete. Although using a name is supported,
                    only types with globally-unique names (e.g. `EObjectType.EVENT`) are supported.
        :param version_control_auto_checkout: Determines if Wwise automatically performs a Checkout source control
                                              operation for affected work units and for the project. Only supported in
                                              Wwise 2023 or above.
        :return: Whether the operation succeeded.
        """
        args = {"object": obj if not isinstance(obj, tuple) else f"{obj[0].get_type_name()}:{obj[1]}",
                **({"autoCheckOutToSourceControl": False} if not version_control_auto_checkout else {})}
        return self._client.call("ak.wwise.core.object.delete", args) is not None
    
    def diff(self, source: GUID | tuple[EObjectType, Name] | ProjectPath,
             target: GUID | tuple[EObjectType, Name] | ProjectPath) -> tuple[tuple[str], tuple[str]]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_diff.html \n
        Compares properties and lists of the source object with those in the target object.
        :param source: The GUID, typed name, or project path of the "source" Wwise object. Although using a name is
                       supported, only types with globally-unique names (e.g. `EObjectType.EVENT`) are supported.
        :param target: The GUID, typed name, or project path of the "target" Wwise object.  Although using a name is
                       supported, only types with globally-unique names (e.g. `EObjectType.EVENT`) are supported.
        :return: Two tuples: the first containing names of properties and references that differ, and the second
                 containing names of lists that differ.
        """
        args = {"source": source if not isinstance(source, tuple) else f"{source[0].get_type_name()}:{source[1]}",
                "target": target if not isinstance(target, tuple) else f"{target[0].get_type_name()}:{target[1]}"}
        results = self._client.call("ak.wwise.core.object.diff", args)
        return results.get("properties", tuple[str]()), results.get("lists", tuple[str]())
    
    def get(self, waql: WaqlQuery | str, returns_and_properties: tuple[EReturnOptions | str, ...] = ()) -> tuple[
        WwiseObjectInfo, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_get.html \n
        Performs a query and returns the data, as specified in the options, for each object in the query result. The
        query should use the WAQL syntax.
        :param waql: A WAQL query, as either a WAQL object or a raw string. Note: **using the `select` statement for
                     objects is valid, but using it with properties (e.g. `Volume`) is invalid.**
        :param returns_and_properties: Additional return options (e.g. `EReturnOptions.WORK_UNIT`) and properties (e.g.
                                       `"Volume"`, `"Pitch"`, etc.). Available properties depend on the type of object
                                       found. For more information and a complete list of properties, check the **Wwise
                                       Objects Reference** page on Audiokinetic's official documentation page. The
                                       requested results will be available in the `other` property of each
                                       `WwiseObjectInfo` instance.
        :return: A collection of `WwiseObjectInfo` instances representing the objects found.
        """
        args = {"waql": str(waql)}  # str conversion needed because of JSON serialization
        
        options = {"return": [*EReturnOptions.get_defaults(), *returns_and_properties]}
        
        objects = self._client.call("ak.wwise.core.object.get", args, options=options)
        objects = objects.get("return", ()) if objects is not None else ()
        
        infos = list[WwiseObjectInfo]()
        for obj in objects:
            info = WwiseObjectInfo.from_dict(obj)
            info.other = {key: value for key, value in obj.items() if key not in EReturnOptions.get_defaults()}
            infos.append(info)
        
        return tuple(infos)
    
    def get_attenuation_curve(self, obj: GUID | Name | ProjectPath, etype: EAttenuationCurveType,
                              platform: GUID | Name = None) -> AttenuationCurve | None:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_getattenuationcurve.html \n
        Gets the specified attenuation curve for a given attenuation object.
        :param obj: The GUID, name, or project path of the Attenuation object to get the curve from.
        :param etype: The type of the attenuation curve to get.
        :param platform: The GUID or unique name of the platform to get the curve from. If unspecified, the platform
                         used will be whichever one is active in the Wwise authoring app or console.
        :return: An AttenuationCurve object describing the curve found.
        """
        args = dict()
        args["object"] = obj if not isinstance(obj, Name) else f"{EObjectType.ATTENUATION.get_type_name()}:{obj}"
        args["curveType"] = etype.value
        if platform is not None:
            args["platform"] = platform
        
        results = self._client.call("ak.wwise.core.object.getAttenuationCurve", args)
        if not results:  # invalid or empty
            return None
        
        points = tuple(GraphPoint2D(Vector2(point["x"], point["y"]),
                                    EnumStatics.from_value(EAttenuationCurveShape, point["shape"]))
                       for point in results["points"])
        usage = EnumStatics.from_value(EAttenuationCurveUsage, results["use"])
        etype = EnumStatics.from_value(EAttenuationCurveType, results["curveType"])
        
        return AttenuationCurve(points, usage, etype)
    
    def get_property_and_reference_names(self, obj: EObjectType | GUID | tuple[EObjectType, Name] | ProjectPath) -> \
            tuple[str]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_getpropertyandreferencenames.html \n
        Retrieves the list of property and reference names for an object.
        :param obj: The type, GUID, typed name, or project path of the Wwise object to get the property and reference
                    names from. Although using a name is supported, only types with globally-unique names (e.g.
                    `EObjectType.EVENT`) are supported. Using the type (`EObjectType`) is recommended over the other
                    options, if possible.
        :return: A tuple of all the property and reference names for the specified object or type.
        """
        match obj:
            case EObjectType():
                args = {"classId": obj.get_class_id()}
            case tuple():
                args = {"object": obj if not isinstance(obj, tuple) else f"{obj[0].get_type_name()}:{obj[1]}"}
            case _:
                args = {"object": obj}
        results = self._client.call("ak.wwise.core.object.getPropertyAndReferenceNames", args)
        return tuple(results["return"]) if results is not None and results.get("return") is not None else ()
    
    def get_property_info(self, obj: EObjectType | GUID | tuple[EObjectType, Name] | ProjectPath,
                          property_name: str) -> PropertyInfo:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_getpropertyinfo.html
        Retrieves information about an object property. Note that this function does not return the value of a
        property. To retrieve the value of a property, refer to `ak.wwise.core.object.get` and Return Options.
        :param obj: The type, GUID, typed name, or project path of the Wwise object to get the property and reference
                    names from. Although using a name is supported, only types with globally-unique names (e.g.
                    `EObjectType.EVENT`) are supported. Using the type (`EObjectType`) is recommended over the other
                    options, if possible.
        :param property_name: The name of the property to retrieve.
        :return: A `PropertyInfo` instance containing information about an object property. That does NOT include the
                 property value.
        """
        match obj:
            case EObjectType():
                obj = obj.get_class_id()
            case tuple():
                obj = f"{obj[0].get_type_name()}:{obj[1]}"
        
        args = {"object": obj, "property": property_name}
        
        info = self._client.call("ak.wwise.core.object.getPropertyInfo", args)
        display = info.get("display", dict())
        supports = info.get("supports", dict())
        return PropertyInfo(info.get("name", ""),
                            display.get("name", ""),
                            info.get("audioEngineId", -1),
                            info.get("default", None),
                            info.get("type", ""),
                            EnumStatics.from_value(ERtpcMode, supports.get("rtpc")),
                            supports.get("unlink", None),
                            supports.get("randomizer", None))
    
    def get_types(self, as_enum: bool = False) -> tuple[dict[str, str | int]] | tuple[EObjectType]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_gettypes.html \n
        Retrieves the list of all object types registered in Wwise's object model. This function returns the equivalent
        of Wwise Objects Reference.
        :param as_enum: If `False`, a WAAPI call is made and a JSON-like dictionary is returned. **If `True`, no WAAPI
                        is actually made, and instead PyWwise will return instances of its `EObjectType` enum type
                        without the need for an actual WAAPI call.** Skipping the WAAPI call and using `EObjectType`
                        objects is available simply as a convenience.
        :return: A `tuple` containing dictionaries - each one describing a single Wwise Object Type - or a `tuple`
                 containing instances of the `EObjectType` enum type (all possible values, essentially).
        """
        if as_enum:
            return tuple[EObjectType](member for member in EObjectType if member != EObjectType.UNKNOWN)
        return tuple(self._client.call("ak.wwise.core.object.getTypes")["return"])
    
    def is_linked(self, obj: GUID | tuple[EObjectType, Name] | ProjectPath, property_name: str,
                  platform: GUID | Name) -> bool | None:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_islinked.html \n
        Indicates whether a property, reference, or object list is bound to a particular platform or to all platforms.
        :param obj: The GUID, typed name, or project path of the object on which to query the link/unlink status.
                    Although using a name is supported, only types with globally-unique names (e.g. `EObjectType.EVENT`)
                    are supported.
        :param property_name: The name of the property for which to query the link/unlink status. Although using a name
                              is supported, only types with globally-unique names (e.g. `EObjectType.EVENT`) are
                              supported.
        :param platform: The GUID or unique name of the platform on which to query the link/unlink status.
        :return: A boolean indicating whether the object is linked for the specified property. If the call failed,
                 this function instead returns None.
        """
        args = {"object": obj if not isinstance(obj, tuple) else f"{obj[0].get_type_name()}:{obj[1]}",
                "property": property_name,
                "platform": platform}
        results = self._client.call("ak.wwise.core.object.isLinked", args)
        return results.get("linked")
    
    def is_property_enabled(self, obj: GUID | tuple[EObjectType, Name] | ProjectPath,
                            property_name: str, platform: GUID | Name) -> bool | None:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_ispropertyenabled.html \n
        Returns true if a property is enabled based on the values of the properties it depends on.
        :param obj: The GUID, typed name, or project path of the object on which to query the link/unlink status.
                    Although using a name is supported, only types with globally-unique names (e.g. `EObjectType.EVENT`)
                    are supported.
        :param property_name: The name of the property for which to query the link/unlink status.
        :param platform: The GUID or unique name of the platform on which to query the link/unlink status.
        :return: Whether a property is enabled based on the values of the properties it depends on.
        """
        args = {"object": obj if not isinstance(obj, tuple) else f"{obj[0].get_type_name()}:{obj[1]}",
                "property": property_name,
                "platform": platform}
        results = self._client.call("ak.wwise.core.object.isPropertyEnabled", args)
        return results.get("return")
    
    def move(self, obj: GUID | tuple[EObjectType, Name] | ProjectPath,
             parent: GUID | tuple[EObjectType, Name] | ProjectPath,
             name_conflict_strategy: ENameConflictStrategy = ENameConflictStrategy.FAIL,
             version_control_auto_checkout: bool = True) -> WwiseObjectInfo | None:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_move.html \n
        Moves an object to the given parent. Returns the moved object.
        :param obj: The GUID, typed name, or project path of the object to move. Although using a name is supported,
                    only types with globally-unique names (e.g. `EObjectType.EVENT`) are supported.
        :param parent: The GUID, typed name, or project path of the object's new parent. Although using a name is supported,
                       only types with globally-unique names (e.g. `EObjectType.EVENT`) are supported.
        :param name_conflict_strategy: The strategy to use in case of a name conflict.
        :param version_control_auto_checkout: Determines if Wwise automatically performs a Checkout source control
                                              operation for affected work units and for the project. Only supported
                                              in Wwise 2023 or above.
        """
        args = {"object": f"{obj[0].get_type_name()}:{obj[1]}" if isinstance(obj, tuple) else obj,
                "parent": f"{parent[0].get_type_name()}:{parent[1]}" if isinstance(parent, tuple) else parent,
                "onNameConflict": name_conflict_strategy,
                **({"autoCheckOutToSourceControl": False} if not version_control_auto_checkout else {})}
        
        results = self._client.call("ak.wwise.core.object.move", args)
        if results is None:
            return None
        
        obj_info = self.get(f"$ from object \"{results.get('id', GUID.get_null())}\" take 1")  # force single match
        return obj_info[0] if len(obj_info) > 0 else None  # WwiseObjectInfo has valid path and type attributes
    
    def paste_properties(self, source: GUID | tuple[EObjectType, Name] | ProjectPath,
                         targets: _Collection[GUID | tuple[EObjectType, Name] | ProjectPath],
                         paste_mode: EPropertyPasteMode = EPropertyPasteMode.REPLACE_ENTIRE, *,
                         property_inclusions: _Collection[str] = None,
                         property_exclusions: _Collection[str] = None) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_pasteproperties.html \n
        Pastes properties, references and lists from one object to any number of target objects. Only those properties,
        references and lists which differ between source and target are pasted. Refer to Wwise Objects Reference for
        more information on the properties, references and lists available on each object type.
        :param source: The GUID, typed name, or project path of the object to copy properties from. Although using a
                       name is supported, only types with globally-unique names (e.g. `EObjectType.EVENT`) are
                       supported.
        :param targets: A collection of GUIDs, typed names, and/or project paths of objects to paste properties into.
                        Although using a name is supported, only types with globally-unique names (e.g.
                        EObjectType.EVENT`) are supported.
        :param paste_mode: Defines how to paste properties into target objects. See options in `EPropertyPasteMode`.
        :param property_inclusions: Array of properties, references and lists to include in the paste operation. When
                                    not specified, all properties, references and lists are included, and the exclusion
                                    defines which ones to exclude.
        :param property_exclusions:	Array of properties, references and lists to exclude from the paste operation. When
                                    not specified, no properties, references and lists are excluded.
        :return: Whether the call succeeded.
        """
        args = {"pasteMode": paste_mode,
                "source": source if not isinstance(source, tuple) else f"{source[0].get_type_name()}:{source[1]}",
                "targets": [target if not isinstance(target, tuple) else f"{target[0].get_type_name()}:{target[1]}"
                            for target in targets]}
        
        if property_inclusions is not None:
            args["inclusion"] = property_inclusions
        elif property_exclusions is not None:
            args["exclusion"] = property_exclusions
        
        return self._client.call("ak.wwise.core.object.pasteProperties", args) is not None
    
    def set(self, operations: ListOrTuple[SetOperation],
            platform: GUID | Name = None,
            on_name_conflict: ENameConflictStrategy = ENameConflictStrategy.FAIL,
            list_mode: EListMode = EListMode.APPEND,
            version_control_auto_add: bool = True) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_set.html \n
        Allows for batch processing of the following operations: Object creation in a child hierarchy, Object creation
        in a list, Setting name, notes, properties and references. See the official Audiokinetic documentation for
        details and complex examples.
        :param operations: A data structure describing the operations to execute. This involves creating and modifying
                           objects.
        :param platform: If targeting a specific platform, you must specify its GUID or unique Name.
        :param on_name_conflict: The strategy to use when solving name conflicts.
        :param list_mode: The strategy to use when an object already exists in a list.
        :param version_control_auto_add: Whether objects should be automatically added to, removed from, and/or
                                         edited in version control. Only supported in Wwise 2023 or above.
        :return: Whether the call succeeded.
        """
        args = {"objects": [operation.dict() for operation in operations],
                "onNameConflict": on_name_conflict,
                "listMode": list_mode,
                **({"platform": platform} if platform is not None else {}),
                **({"autoAddToSourceControl": False} if not version_control_auto_add else {})}
        return self._client.call("ak.wwise.core.object.set", args) is not None
    
    def set_attenuation_curve(self, obj: GUID | Name | ProjectPath,
                              curve_type: EAttenuationCurveType,
                              usage: EAttenuationCurveUsage,
                              points: ListOrTuple[GraphPoint2D],
                              platform: GUID | Name = None):
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_setattenuationcurve.html \n
        Sets the specified attenuation curve for a given attenuation object.
        :param obj: The GUID, name, or project path of the attenuation curve.
        :param curve_type: The attenuation curve type to use.
        :param usage: Defines if the curve has no points, has its own set of points, or uses those of the
                      VolumeDryUsage curve.
        :param points: An array of points defining a curve.
        :param platform: If setting an attenuation curve for a specific platform, you must pass the platform GUID or
                         Name.
        :return: Whether the call succeeded.
        """
        args = {"object": obj if not isinstance(obj, Name) else f"{EObjectType.ATTENUATION.get_type_name()}:{obj[1]}",
                "curveType": curve_type,
                "use": usage,
                "points": [{"x": point.position.x, "y": point.position.y, "shape": point.shape} for point in points],
                **({"platform": platform} if platform is not None else {})}
        return self._client.call("ak.wwise.core.object.setAttenuationCurve", args) is not None
    
    def set_linked(self, obj: GUID | tuple[EObjectType, Name] | ProjectPath, property_name: str,
                   platform: GUID | Name, is_linked: bool) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_setlinked.html \n
        Link or unlink a property/reference or object list to a particular platform.
        :param obj: The GUID, typed name, or project path of the object on which to link or unlink a property or
                    reference. Although using a name is supported, only types with globally-unique names (e.g.
                    `EObjectType.EVENT`) are supported.
        :param property_name: The name of the property to link or unlink.
        :param platform: The platform for which to link or unlink.
        :param is_linked: Whether the object should be linked (`True`) or unlinked (`False`).
        :return: Whether the call succeeded.
        """
        args = {"object": obj if not isinstance(obj, tuple) else f"{obj[0].get_type_name()}:{obj[1]}",
                "property": property_name, "platform": platform, "linked": is_linked}
        return self._client.call("ak.wwise.core.object.setLinked", args) is not None
    
    def set_name(self, obj: GUID | tuple[EObjectType, Name] | ProjectPath, new_name: Name | str) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_setname.html \n
        Renames an object.
        :param obj: The GUID, typed name, or project path of the object for which to set notes. Although using a name
                    is supported, only types with globally-unique names (e.g. `EObjectType.EVENT`) are supported.
        :param new_name: The new name.
        :return: Whether this call succeeded.
        """
        args = {"object": obj if not isinstance(obj, tuple) else f"{obj[0].get_type_name()}:{obj[1]}",
                "value": new_name}
        return self._client.call("ak.wwise.core.object.setName", args) is not None
    
    def set_notes(self, obj: GUID | tuple[EObjectType, Name] | ProjectPath, notes: str) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_setnotes.html \n
        Sets the object's notes.
        :param obj: The GUID, typed name, or project path of the object for which to set notes. Although using a name
                    is supported, only types with globally-unique names (e.g. `EObjectType.EVENT`) are supported.
        :param notes: The new notes.
        :return: Whether this call succeeded.
        """
        args = {"object": obj if not isinstance(obj, tuple) else f"{obj[0].get_type_name()}:{obj[1]}", "value": notes}
        return self._client.call("ak.wwise.core.object.setNotes", args) is not None
    
    def set_property(self, obj: GUID | tuple[EObjectType, Name] | ProjectPath,
                     property_name: str, value: _NoneType | bool | int | float | str,
                     platform: GUID | Name = None) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_setproperty.html \n
        Sets a property value of an object for a specific platform. Refer to Wwise Objects Reference for more
        information on the properties available on each object type. Refer to `ak.wwise.core.object.set_reference` to
        set a reference to an object. Refer to `ak.wwise.core.object.get` to obtain the value of a property for an
        object.
        :param obj: The GUID, typed name, or project path of the object for which to set the specified property.
                    Although using a name is supported, only types with globally-unique names (e.g. `EObjectType.EVENT`)
                    are supported.
        :param property_name: The name of the property to set a new value for.
        :param value: The new value for the property.
        :param platform: The GUID or unique name of the platform for which to set the property.
        :return: Whether this call succeeded.
        """
        args = {"object": obj if not isinstance(obj, tuple) else f"{obj[0].get_type_name()}:{obj[1]}",
                "property": property_name,
                "value": value,
                **({"platform": platform} if platform is not None else {})}
        return self._client.call("ak.wwise.core.object.setProperty", args) is not None
    
    def set_randomizer(self, obj: GUID | tuple[EObjectType, Name] | ProjectPath, property_name: str,
                       enabled: bool, min_value: float = None, max_value: float = None,
                       platform: GUID | Name = None) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_setrandomizer.html \n
        Sets the randomizer values of a property of an object for a specific platform. Refer to Wwise Objects Reference
        for more information on the properties available on each object type.
        :param obj: The GUID, typed name, or project path of the object for which to set the randomizer. Although using
                    a name is supported, only types with globally-unique names (e.g. `EObjectType.EVENT`) are supported.
        :param property_name: The name of the property to set the randomizer for.
        :param enabled: If `True`, the randomizer will be enabled; else, if `False`, it will be disabled.
        :param min_value: Minimum value that the randomizer can offset by. Range: [*,0]
        :param max_value: Maximum value that the randomizer can offset by. Range: [0,*]
        :param platform: The GUID or unique name of the platform for which to set the randomizer.
        :return: Whether this call succeeded.
        """
        args = {"object": obj if not isinstance(obj, tuple) else f"{obj[0].get_type_name()}:{obj[1]}",
                "property": property_name,
                "enabled": enabled,
                **({"min": min_value if min_value <= 0.0 else 0.0} if min_value is not None else {}),
                **({"max": max_value if max_value >= 0.0 else 0.0} if max_value is not None else {}),
                **({"platform": platform} if platform is not None else {})}
        return self._client.call("ak.wwise.core.object.setRandomizer", args) is not None
    
    def set_reference(self, obj: GUID | tuple[EObjectType, Name] | ProjectPath,
                      reference_name: str, value: GUID | tuple[EObjectType, Name] | ProjectPath,
                      platform: GUID | Name = None) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_setreference.html \n
        Sets an object's reference value. Refer to Wwise Objects Reference for more information on the references
        available on each object type.
        :param obj: The GUID, typed name, or project path of the object for which to set the reference. Although using
                    a name is supported, only types with globally-unique names (e.g. `EObjectType.EVENT`) are supported.
        :param reference_name: The name of the reference to set.
        :param value: The new value for the reference. This must be a GUID, typed name, or project path. Although
                          using a name is supported, only types with globally-unique names (e.g. `EObjectType.EVENT`)
                          are supported.
        :param platform: The GUID or unique name of the platform for which to set the reference.
        :return: Whether this call succeeded.
        """
        args = {"object": obj if not isinstance(obj, tuple) else f"{obj[0].get_type_name()}:{obj[1]}",
                "reference": reference_name,
                "value": value if not isinstance(value, tuple) else f"{value[0].get_type_name()}:{value[1]}",
                **({"platform": platform} if platform is not None else {})}
        return self._client.call("ak.wwise.core.object.setReference", args) is not None
    
    def set_state_groups(self, obj: GUID | tuple[EObjectType, Name] | ProjectPath,
                         groups: ListOrTuple[GUID | Name | ProjectPath]) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_setstategroups.html \n
        Sets the State Group objects associated with an object. Note, this will remove any previously associated State
        Group.
        :param obj: The GUID, typed name, or project path of the object for which to set the state groups. Although
                    using a name is supported, only types with globally-unique names (e.g. `EObjectType.EVENT`) are
                    supported.
        :param groups: The state groups to set. This can be any collection containing GUIDs, names, or project paths.
                       It is recommended to pass a `tuple` as the collection.
        :return: Whether this call succeeded.
        """
        groups = [group if not isinstance(group, Name) else f"{EObjectType.STATE_GROUP.get_type_name()}:{group}"
                  for group in groups]
        args = {"object": obj if not isinstance(obj, tuple) else f"{obj[0].get_type_name()}:{obj[1]}",
                "stateGroups": groups}
        return self._client.call("ak.wwise.core.object.setStateGroups", args) is not None
    
    def set_state_properties(self, obj: GUID | tuple[EObjectType, Name] | ProjectPath,
                             properties: ListOrTuple[str]) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_object_setstateproperties.html \n
        Set the state properties of an object. Note, this will remove any previous state property, including the
        default ones.
        :param obj: The GUID, typed name, or project path of the object for which to set the state groups. Although
            using a name is supported, only types with globally-unique names (e.g. `EObjectType.EVENT`) are
            supported.
        :param properties: An array containing the names of the state properties to set.
        :return: Whether this call succeeded.
        """
        args = {"object": obj if not isinstance(obj, tuple) else f"{obj[0].get_type_name()}:{obj[1]}",
                "stateProperties": properties}
        return self._client.call("ak.wwise.core.object.setStateProperties", args) is not None



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.enums import ERoomVerbFilterType, ERoomVerbInsertPosition
from pywwise.objects.abc import WwiseObject


class WwiseRoomVerb(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_effect_wwise_roomverb.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.WWISE_ROOMVERB`.
    """
    centre_level = WwiseProperty[float]("CenterLevel", float)
    dc_filter_cutoff_frequency = WwiseProperty[float]("DCFilterCutFreq", float)
    decay_time = WwiseProperty[float]("DecayTime", float)
    density = WwiseProperty[float]("Density", float)
    density_delay_max = WwiseProperty[float]("DensityDelayMax", float)
    density_delay_min = WwiseProperty[float]("DensityDelayMin", float)
    density_delay_rdm_percentage = WwiseProperty[float]("DensityDelayRdmPerc	", float)
    diffusion = WwiseProperty[float]("Diffusion", float)
    diffusion_delay_max = WwiseProperty[float]("DiffusionDelayMax", float)
    diffusion_delay_rdm_percentage = WwiseProperty[float]("DiffusionDelayRdmPerc", float)
    diffusion_delay_scale_percentage = WwiseProperty[float]("DiffusionDelayScalePerc", float)
    dry_level = WwiseProperty[float]("DryLevel", float)
    early_reflections_front_back_delay = WwiseProperty[float]("ERFrontBackDelay", float)
    early_reflections_level = WwiseProperty[float]("ERLevel", float)
    early_reflections_pattern = WwiseProperty[int]("ERPattern", int)
    enable_early_reflections = WwiseProperty[bool]("EnableEarlyReflections", bool)
    enable_tone_controls = WwiseProperty[bool]("EnableToneControls", bool)
    filter1_curve = WwiseProperty[ERoomVerbFilterType]("Filter1Curve", ERoomVerbFilterType)
    filter1_frequency = WwiseProperty[float]("Filter1Freq", float)
    filter1_gain = WwiseProperty[float]("Filter1Gain", float)
    filter1_insert_position = WwiseProperty[ERoomVerbInsertPosition]("Filter1InsertPos", ERoomVerbInsertPosition)
    filter1_q = WwiseProperty[float]("Filter1Q", float)
    filter2_curve = WwiseProperty[ERoomVerbFilterType]("Filter2Curve", ERoomVerbFilterType)
    filter2_frequency = WwiseProperty[float]("Filter2Freq", float)
    filter2_gain = WwiseProperty[float]("Filter2Gain", float)
    filter2_insert_position = WwiseProperty[ERoomVerbInsertPosition]("Filter2InsertPos", ERoomVerbInsertPosition)
    filter2_q = WwiseProperty[float]("Filter2Q", float)
    filter3_curve = WwiseProperty[ERoomVerbFilterType]("Filter3Curve", ERoomVerbFilterType)
    filter3_frequency = WwiseProperty[float]("Filter3Freq", float)
    filter3_gain = WwiseProperty[float]("Filter3Gain", float)
    filter3_insert_position = WwiseProperty[ERoomVerbInsertPosition]("Filter3InsertPos", ERoomVerbInsertPosition)
    filter3_q = WwiseProperty[float]("Filter3Q", float)
    front_level = WwiseProperty[float]("FrontLevel", float)
    high_frequency_damping = WwiseProperty[float]("HFDamping", float)
    input_centre_level = WwiseProperty[float]("InputCenterLevel", float)
    input_lfe_level = WwiseProperty[float]("InputLFELevel", float)
    lfe_level = WwiseProperty[float]("LFELevel", float)
    pre_delay = WwiseProperty[float]("PreDelay", float)
    quality = WwiseProperty[int]("Quality", int)
    rear_level = WwiseProperty[float]("RearLevel", float)
    reverb_level = WwiseProperty[float]("ReverbLevel", float)
    reverb_unit_input_delay = WwiseProperty[float]("ReverbUnitInputDelay", float)
    reverb_unit_input_delay_rmd_percentage = WwiseProperty[float]("ReverbUnitInputDelayRmdPerc", float)
    room_shape = WwiseProperty[float]("RoomShape", float)
    room_shape_max = WwiseProperty[float]("RoomShapeMax", float)
    room_shape_min = WwiseProperty[float]("RoomShapeMin", float)
    room_size = WwiseProperty[float]("RoomSize", float)
    stereo_width = WwiseProperty[float]("StereoWidth", float)



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.objects.abc import WwiseObject


class WwiseSilence(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_source_wwise_silence.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.SOURCE_PLUGIN`.
    """
    length = WwiseProperty[float]("Length", float)
    length_max = WwiseProperty[float]("LengthMax", float)
    length_min = WwiseProperty[float]("LengthMin", float)



# Copyright 2024 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from waapi import WaapiClient as _WaapiClient

from pywwise.aliases import ListOrTuple
from pywwise.enums import EActionOnEventType, EFadeCurve, EMultiPositionType, ESpeakerBitMask
from pywwise.primitives import GameObjectID, GUID, Name, PlayingID, ProjectPath, ShortID
from pywwise.structs import AuxSendValue, Vector3


class SoundEngine:
    """ak.soundengine"""
    
    def __init__(self, client: _WaapiClient):
        """
        Constructor.
        :param client: The WAAPI client to use.
        """
        self._client = client
    
    def execute_action_on_event(self, event: Name | ShortID | GUID, action_type: EActionOnEventType,
                                game_object: GameObjectID, transition_duration: int = 0,
                                fade_curve: EFadeCurve = EFadeCurve.LINEAR) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_soundengine_executeactiononevent.html \n
        Executes an action on all nodes that are referenced in the specified event in a Play action.
        :param event: The name, short ID, or GUID of the event.
        :param action_type: The type of action (e.g. Play).
        :param game_object: The ID of the game object.
        :param transition_duration: The fade duration in milliseconds.
        :param fade_curve: The curve to use for the fade.
        :return: Whether this call succeeded.
        """
        args = {"event": event, "actionType": action_type, "gameObject": game_object,
                "transitionDuration": transition_duration, "fadeCurve": fade_curve}
        result = self._client.call("ak.soundengine.executeActionOnEvent", args)
        return bool(result) if result is not None else False
    
    def get_state(self, state_group: Name | ShortID | GUID | ProjectPath) -> tuple[str, str]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_soundengine_getstate.html \n
        Gets the current state of a State Group. When using setState just prior to getState, allow a brief delay (no
        more than 10ms) for the information to update in the sound engine.
        :param state_group: The name, short ID, GUID, or project path of the State Group.
        :return: The name and GUID of the active State value. If invalid, the values will be -1 and "".
        """
        if isinstance(state_group, Name):
            state_group = f"StateGroup:{state_group}"
        elif isinstance(state_group, ShortID):
            state_group = f"Global:{state_group}"
        args = {"stateGroup": state_group}
        results = self._client.call("ak.soundengine.getState", args).get("return")
        return results.get("name"), results.get("id")
    
    def get_switch(self, switch_group: Name | ShortID | GUID | ProjectPath,
                   game_object: GameObjectID) -> tuple[str, str]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_soundengine_getswitch.html \n
        Gets the current state of a Switch Group for a given Game Object.
        :param switch_group: The name, short ID, or GUID of the Switch Group.
        :param game_object: The ID of the game object. Switches are always encapsulated by game objects.
        :return: The name and GUID of the active Switch value.
        """
        if isinstance(switch_group, Name):
            switch_group = f"SwitchGroup:{switch_group}"
        elif isinstance(switch_group, ShortID):
            switch_group = f"Global:{switch_group}"
        args = {"switchGroup": switch_group, "gameObject": game_object}
        results = self._client.call("ak.soundengine.getSwitch", args).get("return")
        return results.get("name"), results.get("id")
    
    def load_bank(self, sound_bank: Name | ShortID | GUID) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_soundengine_loadbank.html \n
        Load a SoundBank. See `AK::SoundEngine::LoadBank`.
        :param sound_bank: The name or short ID of the bank.
        :return: Whether the operation worked. True does not mean the bank was not loaded prior.
        """
        args = {"soundBank": sound_bank}
        return self._client.call("ak.soundengine.loadBank", args) is not None
    
    def post_event(self, event: Name | ShortID | GUID,
                   game_object: GameObjectID = GameObjectID.get_transport()) -> int:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_soundengine_postevent.html \n
        Asynchronously post an Event to the sound engine (by event ID). See `AK::SoundEngine::PostEvent`.
        :param event: The name, short ID, or GUID of the event to post.
        :param game_object: The ID of the game object the event should be posted on.
        :return: The Playing ID of the event. If the call failed, this will be -1.
        """
        args = {"event": event, "gameObject": game_object}
        results = self._client.call("ak.soundengine.postEvent", args)
        return results.get("return") if results is not None else -1
    
    def post_msg_monitor(self, message: str) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_soundengine_postmsgmonitor.html \n
        Display a message in the Profiler's Capture Log view.
        :param message: The message to display.
        :return: Whether this operation worked.
        """
        args = {"message": message}
        return self._client.call("ak.soundengine.postMsgMonitor", args) is not None
    
    def post_trigger(self, trigger: Name | ShortID | GUID, game_object: GameObjectID = None) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_soundengine_posttrigger.html \n
        Posts the specified Trigger. See `AK::SoundEngine::PostTrigger`.
        :param trigger: The name, short ID, or GUID of the trigger to post.
        :param game_object: The ID of the game object on which the trigger should be posted. If unspecified, the trigger
                            will be posted globally.
        :return: Whether this operation worked. True does not mean IDs were all valid.
        """
        args = {"trigger": trigger}
        if game_object is not None:
            args["gameObject"] = game_object
        return self._client.call("ak.soundengine.postTrigger", args) is not None
    
    def register_game_obj(self, obj_id: GameObjectID, obj_name: Name) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_soundengine_registergameobj.html \n
        Register a game object. Registering a game object twice does nothing. Unregistering it once unregisters
        it no matter how many times it has been registered. See `AK::SoundEngine::RegisterGameObj`.
        :param obj_id: The ID of the game object to register.
        :param obj_name: The name of the game object to register. This is for monitoring purposes only.
        :return: Whether the call succeeded.
        """
        args = {"gameObject": obj_id, "name": obj_name}
        return self._client.call("ak.soundengine.registerGameObj", args) is not None
    
    def reset_rtpc_value(self, rtpc: Name | GUID | ShortID, game_object: GameObjectID) -> bool:
        """
        https://www.audiokinetic.com/library/2023.1.4_8496/?source=SDK&id=ak_soundengine_resetrtpcvalue.html \n
        Resets the value of a real-time parameter control to its default value, as specified in the Wwise
        project. See `AK::SoundEngine::ResetRTPCValue`. For a global operation, use GameObjectID.get_global().
        :param rtpc: The name, GUID, or short ID of the RTPC that should be reset to its default value.
        :param game_object: The game object ID (integer).
        :return: Whether the call succeeded. True does not mean a valid RTPC was necessarily reset.
        """
        args = {"rtpc": rtpc, "gameObject": game_object}
        return self._client.call("ak.soundengine.resetRTPCValue", args) is not None
    
    def seek_on_event(self, event: Name | GUID | ShortID, game_object: GameObjectID, position: int | float,
                      seek_to_nearest_marker: bool = False, playing_id: PlayingID = 0) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_soundengine_seekonevent.html \n
        Seeks inside all playing objects that are referenced in Play Actions of the specified Event. See
        `AK::SoundEngine::SeekOnEvent`.
        :param event: The name, GUID, or short ID of the event to seek on.
        :param game_object: The ID of the GameObject that owns (encapsulates) the event. Note: this API uses uint32;
                            this type has a maximum valid value is `4294967295`, therefore the default Transport and
                            Global IDs are NOT supported.
        :param position: The position where to seek. Note: `int` implies milliseconds; `float` implies percentage.
        :param seek_to_nearest_marker: If true, the final seeking position is made equal to the nearest marker.
        :param playing_id: The playing ID for which the seek if to be applied.
        :return: Whether the call succeeded. True does not mean the Seek action necessarily worked.
        """
        seek_type = "position" if isinstance(position, int) else "percent"
        args = {"event": event, "gameObject": game_object, seek_type: position,
                "seekToNearestMarker": seek_to_nearest_marker, "playingId": playing_id}
        return self._client.call("ak.soundengine.seekOnEvent", args) is not None
    
    def set_default_listeners(self, listeners: ListOrTuple[GameObjectID]) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_soundengine_setdefaultlisteners.html \n
        Sets the default active listeners for all subsequent game objects that are registered. See
        `AK::SoundEngine::SetDefaultListeners`.
        :param listeners: The array of listener game object IDs. Game objects must have been previously registered.
                          Duplicates are ignored.
        :return: Whether the call succeeded. True does not mean the informed IDs were all valid.
        """
        listeners = list(dict.fromkeys(listeners))
        args = {"listeners": listeners, "numListeners": len(listeners)}
        return self._client.call("ak.soundengine.setDefaultListeners", args) is not None
    
    def set_game_object_aux_send_values(self, game_obj: GameObjectID,
                                        aux_send_values: ListOrTuple[AuxSendValue]) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_soundengine_setgameobjectauxsendvalues.html \n
        Sets the Auxiliary Buses to route the specified game object. See `AK::SoundEngine::SetGameObjectAuxSendValues`.
        :param game_obj: The game object ID associated with this operation.
        :param aux_send_values: The aux send values to set.
        :return: Whether the call succeeded. True does not mean the volume necessarily changed (e.g. invalid IDs).
        """
        args = {"gameObject": game_obj, "auxSendValues": []}
        for aux_send_value in aux_send_values:
            args["auxSendValues"].append({"listener": aux_send_value.listener,
                                          "auxBus": aux_send_value.aux_bus,
                                          "controlValue": aux_send_value.control_value})
        return self._client.call("ak.soundengine.setGameObjectOutputBusVolume", args) is not None
    
    def set_game_object_output_bus_volume(self, emitter: GameObjectID, listener: GameObjectID,
                                          control_value: float) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_soundengine_setgameobjectoutputbusvolume.html \n
        Set the output bus volume (direct) to be used for the specified game object. See
        `AK::SoundEngine::SetGameObjectOutputBusVolume`.
        :param emitter: The ID of the emitter game object.
        :param listener: The ID of the listener game object.
        :param control_value: A multiplier where 0 means silence and 1 means no change. Therefore, values between 0 and
                              1 attenuate the sound, and values greater than 1 amplify it.
        :return: Whether the call succeeded. True does not mean the volume necessarily changed (e.g. invalid IDs).
        """
        args = {"emitter": emitter, "listener": listener, "controlValue": control_value}
        return self._client.call("ak.soundengine.setGameObjectAuxSendValues", args) is not None
    
    def set_listeners(self, emitter: GameObjectID, listeners: ListOrTuple[GameObjectID]) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_soundengine_setlisteners.html \n
        Sets a single game object's active listeners. By default, all new game objects have no listeners active,
        but this behavior can be overridden with `SetDefaultListeners()`. Inactive listeners are not computed. See
        `AK::SoundEngine::SetListeners`.
        :param emitter: The ID of the emitter Game Object to set listeners for.
        :param listeners: The ID of the listeners being set. Duplicates are ignored.
        :return: Whether the call succeeded. Note: passing unregistered IDs will cause no change, but the function will
                 still return True.
        """
        listeners = list(dict.fromkeys(listeners))
        args = {"emitter": emitter, "listeners": list(listeners)}
        return self._client.call("ak.soundengine.setListeners", args) is not None
    
    def set_listener_spatialization(self, listener: GameObjectID, spatialized: bool, channel_config: ESpeakerBitMask,
                                    volume_offsets: ListOrTuple[float]) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_soundengine_setlistenerspatialization.html \n
        Sets a listener's spatialization parameters. This lets you define listener-specific volume offsets for
        each audio channel. See `AK::SoundEngine::SetListenerSpatialization`.
        :param listener: The ID of the listener.
        :param spatialized: Whether the listener should be spatialized.
        :param channel_config: The channel configuration (e.g. 5.1).
        :param volume_offsets: The volume offsets. Make sure this matches the amount of channels.
        :return: Whether the call succeeded. True does not mean the arguments where all valid.
        """
        args = {"listener": listener, "spatialized": spatialized,
                "channelConfig": channel_config.value, "volumeOffsets": volume_offsets}
        return self._client.call("ak.soundengine.setListenerSpatialization", args) is not None
    
    def set_multiple_positions(self, game_obj: GameObjectID, orientation_front: ListOrTuple[Vector3],
                               orientation_top: ListOrTuple[Vector3], position: ListOrTuple[Vector3],
                               multi_position_type: EMultiPositionType) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_soundengine_setmultiplepositions.html \n
        Sets multiple positions for a single game object. Setting multiple positions for a single game object is
        a way to simulate multiple emission sources while using the resources of only one voice. This can be used
        to simulate wall openings, area sounds, or multiple objects emitting the same sound in the same area. See
        `AK::SoundEngine::SetMultiplePositions`.
        :param game_obj: The ID of the game object to set the position for.
        :param orientation_front: The orientation (front).
        :param orientation_top: The orientation (top).
        :param position: The position to set.
        :param multi_position_type: How to set up the multiple positions.
        :raise ValueError: If the lengths of orientation_front, orientation_top, and position are not equal.
        :return: Whether the call succeeded.
        """
        if len({orientation_front, orientation_top, position}) != 1:  # 1 means different lengths, in this context
            raise ValueError("The amount of values in `orientation_front`, `orientation_top`, "
                             "and `position` must be equal.")
        
        args = {"gameObject": game_obj,
                "multiPositionType": multi_position_type,
                "positions": [{"orientationFront": orientation_front[i],
                               "orientationTop": orientation_top[i],
                               "position": position[i]} for i in range(len(position))]}
        
        return self._client.call("ak.soundengine.setMultiplePositions", args) is not None
    
    def set_object_obstruction_and_occlusion(self, emitter: GameObjectID, listener: GameObjectID,
                                             obstruction_level: float, occlusion_level: float) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_soundengine_setobjectobstructionandocclusion.html \n
        Set a game object's obstruction and occlusion levels. This function is used to affect how an object
        should be heard by a specific listener. See `AK::SoundEngine::SetObjectObstructionAndOcclusion`.
        :param emitter: The ID of the emitter.
        :param listener: The ID of the listener.
        :param obstruction_level: The level of obstruction. Range: 0.0f to 1.0f.
        :param occlusion_level: The level of occlusion. Range: 0.0f to 1.0f.
        :return: Whether the call succeeded. True does not mean the arguments where all valid.
        """
        args = {"emitter": emitter, "listener": listener, "obstructionLevel": obstruction_level,
                "occlusionLevel": occlusion_level}
        return self._client.call("ak.soundengine.setObjectObstructionAndOcclusion", args) is not None
    
    def set_position(self, game_obj: GameObjectID, orientation_front: Vector3, orientation_top: Vector3,
                     position: Vector3) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_soundengine_setposition.html \n
        Sets the position of a game object. See `AK::SoundEngine::SetPosition`.
        :param game_obj: The ID of the game object to set the position for.
        :param orientation_front: The orientation (front).
        :param orientation_top: The orientation (top).
        :param position: The position to set.
        :return: Whether the call succeeded.
        """
        args = {"gameObject": game_obj, "position": dict()}
        args["position"]["orientationFront"] = {"x": orientation_front.x, "y": orientation_front.y,
                                                "z": orientation_front.z}
        args["position"]["orientationTop"] = {"x": orientation_top.x, "y": orientation_top.y, "z": orientation_top.z}
        args["position"]["position"] = {"x": position.x, "y": position.y, "z": position.z}
        return self._client.call("ak.soundengine.setPosition", args) is not None
    
    def set_rtpc_value(self, rtpc: GUID | Name | ShortID, value: float,
                       game_obj: GameObjectID = GameObjectID.get_global()) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_soundengine_setrtpcvalue.html \n
        Sets the value of a real-time parameter control. See `AK::SoundEngine::SetRTPCValue`.
        :param rtpc: The GUID, name, or short ID of the RTPC to set a new value for.
        :param value: The new value of the RTPC.
        :param game_obj: If specified, the RTPC will be set locally. Else, it will be set globally.
        :return: Whether this call succeeded. True does not necessarily mean the RTPC, value, and game object were valid.
        """
        args = {"rtpc": rtpc, "value": value}
        if game_obj is not None:
            args["gameObject"] = game_obj
        return self._client.call("ak.soundengine.setRTPCValue", args) is not None
    
    def set_scaling_factor(self, game_obj: GameObjectID, attenuation_scaling_factor: float) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_soundengine_setscalingfactor.html \n
        Sets the scaling factor of a game object. You can modify the attenuation computations on this game object
        to simulate sounds with a larger or smaller affected areas. See `AK::SoundEngine::SetScalingFactor`.
        :param game_obj: The ID of the game object.
        :param attenuation_scaling_factor: The scaling factor, where 1 means 100%, 0.5 means 50%, 2 means 200%, and so on.
        :return: Whether the call succeeded. True does not mean the arguments where all valid.
        """
        args = {"gameObject": game_obj, "attenuationScalingFactor": attenuation_scaling_factor}
        return self._client.call("ak.soundengine.setScalingFactor", args) is not None
    
    def set_state(self, state_group: GUID | Name | ShortID, state_value: GUID | Name | ShortID) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_soundengine_setstate.html \n
        Sets the State of a State Group. See `AK::SoundEngine::SetState`.
        :param state_group: The GUID, Name, or Short ID of the state group.
        :param state_value: The GUID, Name, or Short ID of the state to set.
        :return: Whether the call succeeded. True does not mean the arguments where all valid.
        """
        args = {"stateGroup": state_group, "state": state_value}
        return self._client.call("ak.soundengine.setState", args) is not None
    
    def set_switch(self, switch_group: GUID | Name | ShortID, switch_value: GUID | Name | ShortID,
                   game_obj: GameObjectID) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_soundengine_setswitch.html \n
        Sets the State of a Switch Group. See `AK::SoundEngine::SetSwitch`.
        :param switch_group: The GUID, Name, or Short ID of the switch group.
        :param switch_value: The GUID, Name, or Short ID of the switch to set.
        :param game_obj: The ID of the game object.
        :return: Whether the call succeeded. True does not mean the arguments where all valid.
        """
        args = {"switchGroup": switch_group, "switchState": switch_value, "gameObject": game_obj}
        return self._client.call("ak.soundengine.setSwitch", args) is not None
    
    def stop_all(self, game_obj: GameObjectID):
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_soundengine_stopall.html \n
        Stop playing the current content associated to the specified game object ID. If no game object is
        specified, all sounds are stopped. See `AK::SoundEngine::StopAll`.
        :param game_obj: The ID of the game object.
        :return: Whether the call succeeded. True does not mean the arguments where all valid.
        """
        args = {"gameObject": game_obj}
        return self._client.call("ak.soundengine.stopAll", args) is not None
    
    def stop_playing_id(self, playing_id: PlayingID, transition_duration: int, fade_curve: int):
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_soundengine_stopplayingid.html \n
        Stops the current content, associated to the specified playing ID, from playing. See
        `AK::SoundEngine::StopPlayingID`.
        :param playing_id: The Playing ID to be stopped.
        :param transition_duration: The fade duration for the transition (ms).
        :param fade_curve: The curve type to be used in the transition. Uses values from AkCurveInterpolation. Range: [0,9].
        :return: Whether the call succeeded. True does not mean the arguments where all valid.
        """
        args = {"playingId": playing_id, "transitionDuration": transition_duration, "fadeCurve": fade_curve}
        return self._client.call("ak.soundengine.stopPlayingID", args) is not None
    
    def unload_bank(self, sound_bank: Name | ShortID | GUID) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_soundengine_unloadbank.html \n
        Unload a SoundBank. See `AK::SoundEngine::UnloadBank`.
        :param sound_bank: The name, short ID, or GUID of the bank to unload.
        :return: Whether the operation succeeded. True does not mean the bank was previously loaded.
        """
        args = {"soundBank": sound_bank}
        return self._client.call("ak.soundengine.unloadBank", args) is not None
    
    def unregister_game_obj(self, obj_id: int) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_soundengine_unregistergameobj.html \n
        Unregisters a game object. Registering a game object twice does nothing. Unregistering it once
        unregisters it no matter how many times it has been registered. Unregistering a game object while it is
        in use is allowed, but the control over the parameters of this game object is lost. For example,
        say a sound associated with this game object is a 3D moving sound. It stops moving when the game object
        is unregistered, and there is no way to regain control over the game object. See
        `AK::SoundEngine::UnregisterGameObj`.
        :param obj_id: The ID of the game object to unregister.
        :return: Whether the call succeeded. True does not mean the arguments where all valid.
        """
        args = {"gameObject": obj_id}
        return self._client.call("ak.soundengine.unregisterGameObj", args) is not None



# Copyright 2024 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from simplevent import RefEvent as _RefEvent
from waapi import WaapiClient as _WaapiClient

from pywwise.aliases import ListOrTuple, SystemPath
from pywwise.decorators import callback
from pywwise.enums import (EAudioObjectOptions, EBusOptions, EDataTypes, EObjectType, EReturnOptions, ETimeCursor,
                           EVoicePipelineReturnOptions)
from pywwise.primitives import GameObjectID, GUID, Name, ProjectPath, ShortID
from pywwise.structs import (ActiveRTPCInfo, AudioObjectInfo, AudioObjectMetadata, BusPipelineInfo, CPUStatisticsInfo,
                             GameObjectRegistrationData, LoadedMediaInfo, PerformanceMonitorCounterInfo,
                             PlayingVoiceProperties, StreamObjectInfo, VoiceContributionHierarchy,
                             VoiceContributionParameter, VoiceInspectorContribution, WwiseObjectInfo)
from pywwise.waapi.ak.wwise.core.capture_log import CaptureLog as _CaptureLog


class Profiler:
    """ak.wwise.core.profiler"""
    
    def __init__(self, client: _WaapiClient):
        """
        Constructor.
        :param client: The WAAPI client to use.
        """
        self._client = client
        self.capture_log = _CaptureLog(client)
        
        self.game_object_registered = _RefEvent(int, GameObjectID, Name)
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_profiler_gameobjectregistered.html
        \nSent when a game object has been registered.
        \n**Event Data**:
        \n- Time of registration. Elapsed time in milliseconds since the initialization of the soundengine.
        \n- The game object ID for the entry.
        \n- The game object name for the entry.
        """
        
        self._game_object_registered = self._client.subscribe("ak.wwise.core.profiler.gameObjectRegistered",
                                                              self._on_game_object_registered)
        
        self.game_object_reset = _RefEvent()
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_profiler_gameobjectreset.html
        \nSent when the game objects have been reset, such as closing a connection to a game while profiling.
        """
        
        self._game_object_reset = self._client.subscribe("ak.wwise.core.profiler.gameObjectReset",
                                                         self._on_game_object_reset)
        
        self.game_object_unregistered = _RefEvent(int, GameObjectID, Name)
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_profiler_gameobjectunregistered.html
        \nSent when a game object has been unregistered.
        \n**Event Data**:
        \n- Time of un-registration. Elapsed time in milliseconds since the initialization of the soundengine.
        \n- The game object ID for the entry.
        \n- The game object name for the entry.
        """
        
        self._game_object_unregistered = self._client.subscribe("ak.wwise.core.profiler.gameObjectUnregistered",
                                                                self._on_game_object_unregistered)
        
        # The return options below are needed so we can retrieve information about Switch and States, plus their Groups.
        change_args = {"return": [EReturnOptions.GUID, EReturnOptions.NAME, EReturnOptions.TYPE, EReturnOptions.PATH]}
        
        self.state_changed = _RefEvent(WwiseObjectInfo, WwiseObjectInfo)
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_profiler_statechanged.html
        \nSent when a state group state has been changed. Does not require the profiler capture log to be started.
        \n**Event Data**:
        \n-A WwiseObjectInfo instance containing information about the State Group where the change happened.
        \n-A WwiseObjectInfo instance containing information about the new active State.
        """
        
        self._state_changed = self._client.subscribe("ak.wwise.core.profiler.stateChanged",
                                                     self._on_state_changed, change_args)
        
        self.switch_changed = _RefEvent(WwiseObjectInfo, WwiseObjectInfo, GameObjectID)
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_profiler_switchchanged.html
        \nSent when a switch group state has been changed. Does not require the profiler capture log to be started.
        \n**Event Data**:
        \n-A WwiseObjectInfo instance containing information about the Switch Group where the change happened.
        \n-A WwiseObjectInfo instance containing information about the new active Switch.
        \n-The ID of the game object on which the change happened.
        """
        
        self._switch_changed = self._client.subscribe("ak.wwise.core.profiler.switchChanged",
                                                      self._on_switch_changed, change_args)
    
    @callback
    def _on_game_object_registered(self, event: _RefEvent, **kwargs):
        """
        Callback function for the `gameObjectRegistered` event.
        :param event: The event to broadcast.
        :param kwargs: The event data.
        """
        event(kwargs["time"], GameObjectID(kwargs["gameObjectId"]), Name(kwargs["gameObjectName"]))
    
    @callback
    def _on_game_object_reset(self, event: _RefEvent):
        """
        Callback function for the `gameObjectReset` event.
        :param event: The event to broadcast.
        """
        event()
    
    @callback
    def _on_game_object_unregistered(self, event: _RefEvent, **kwargs):
        """
        Callback function for the `gameObjectUnregistered` event.
        :param event: The event to broadcast.
        :param kwargs: The event data.
        """
        event(kwargs["time"], GameObjectID(kwargs["gameObjectId"]), Name(kwargs["gameObjectName"]))
    
    @callback
    def _on_state_changed(self, event: _RefEvent, **kwargs):
        """
        Callback function for the `stateChanged` event.
        :param event: The event to broadcast.
        :param kwargs: The event data.
        """
        group = WwiseObjectInfo.from_dict(kwargs["stateGroup"])
        value = WwiseObjectInfo.from_dict(kwargs["state"])
        event(group, value)
    
    @callback
    def _on_switch_changed(self, event: _RefEvent, **kwargs):
        """
        Callback function for the `switchChanged` event.
        :param event: The event to broadcast.
        :param kwargs: The event data.
        """
        group = WwiseObjectInfo.from_dict(kwargs["switchGroup"])
        value = WwiseObjectInfo.from_dict(kwargs["switch"])
        event(group, value, kwargs["gameObjectID"])
    
    def enable_profiler_data(self, data_types: ListOrTuple[tuple[EDataTypes, bool]]) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_profiler_enableprofilerdata.html \n
        Specifies the type of data you want to capture. Overrides the user's profiler settings.
        :param data_types: An array of tuples. Each tuple contains the data type you want to capture via an enum and
                           if profiler capture will be enabled for said data type (Defaults to true).
        :return: It returns a bool indicating whether the call was successful or not.
        """
        data_types = list(dict.fromkeys(data_types))
        args = {"dataTypes": list()}
        
        for data_type in data_types:
            args["dataTypes"].append({"dataType": data_type[0], "enabled": data_type[1]})
        
        return self._client.call("ak.wwise.core.profiler.enableProfilerData", args) is not None
    
    def get_audio_objects(self, time: ETimeCursor | int, bus_pipeline_id: int = None,
                          returns: ListOrTuple[EAudioObjectOptions] = None) -> tuple[AudioObjectInfo, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_profiler_getaudioobjects.html \n
        Retrieves the Audio Objects at a specific profiler capture time.
        :param time: Time in milliseconds to query for Audio Objects, or a Time Cursor from which to acquire the time.
                     This parameter can have 2 possible values: int or ETimeCursor. The int is the time to query.
                     The ETimeCursor can have two values: user or capture. The User Time Cursor is the one that can
                     be manipulated by the user, while the Capture Time Cursor represents the latest time of the current
                     capture.
        :param bus_pipeline_id: Unsigned Integer 32-bit. Range: [0,4294967295] The pipeline ID of a Bus instance to get.
        :param returns: Members to return for each Audio Object. Defaults to Audio Object ID, Bus Pipeline ID,
                        Instigator Pipeline ID and Effect Class ID.
        :return: For each profiled audio object, an AudioObjectInfo containing an object's ID, pipeline ID,
                 instigator pipeline ID, effect class ID and a dictionary containing additional information
                 (requested via `return_options`). When accessing the values in the dictionary, use the
                 EAudioObjectOptions enum as the keys. If this function call fails, an empty tuple is returned.
        """
        args = {"time": time}
        
        if bus_pipeline_id is not None:
            args["busPipelineId"] = bus_pipeline_id
        
        returns = list(returns) if returns is not None else list[EAudioObjectOptions]()
        returns.extend((EAudioObjectOptions.AUDIO_OBJECT_ID, EAudioObjectOptions.BUS_PIPELINE_ID,
                        EAudioObjectOptions.INSTIGATOR_PIPELINE_ID, EAudioObjectOptions.EFFECT_CLASS_ID,))
        options = {"return": list(dict.fromkeys(returns))}  # Remove duplicates
        
        results = self._client.call("ak.wwise.core.profiler.getAudioObjects", args, options=options)
        results = results.get("return", list[dict]())
        
        objects = list[AudioObjectInfo]()
        
        for result in results:
            info = AudioObjectInfo(result[EAudioObjectOptions.AUDIO_OBJECT_ID],
                                   result[EAudioObjectOptions.BUS_PIPELINE_ID],
                                   result[EAudioObjectOptions.INSTIGATOR_PIPELINE_ID],
                                   result[EAudioObjectOptions.EFFECT_CLASS_ID],
                                   {k: v for k, v in result.items()
                                    if k not in returns and k != EAudioObjectOptions.METADATA})
            if result.get(EAudioObjectOptions.METADATA) is not None:
                info.other[EAudioObjectOptions.METADATA] = list()
            for meta in result.get(EAudioObjectOptions.METADATA, ()):
                info.other[EAudioObjectOptions.METADATA].append(
                    AudioObjectMetadata(meta.get("metadataClassID", -1),
                                        meta.get("sourceShortID", ShortID.get_null()),
                                        meta.get("metadataName", Name.get_null()),
                                        meta.get("sourceID", GUID.get_null()),
                                        meta.get("sourceName", Name.get_null())))
            objects.append(info)
        
        return tuple(objects)
    
    def get_busses(self, time: ETimeCursor | int, bus_pipeline_id: int = None,
                   returns: ListOrTuple[EBusOptions] = None) -> tuple[BusPipelineInfo, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_profiler_getbusses.html \n
        Retrieves the busses at a specific profiler capture time.
        :param time: Time in milliseconds to query for a bus pipeline, or a Time Cursor from which to acquire the time.
                     This parameter can have 2 possible values: int or ETimeCursor. The int is the time to query.
                     The ETimeCursor can have two values: user or capture. The User Time Cursor is the one that can
                     be manipulated by the user, while the Capture Time Cursor represents the latest time of the current
                     capture.
        :param bus_pipeline_id: Unsigned Integer 32-bit. Range: [0,4294967295] The pipeline ID of a single bus instance
                                to get.
        :param returns: Members to return for each bus. If `None` the default returns are used: Pipeline ID, GameObject
                        ID, and Object GUID.
        :return: For each profiled bus pipeline, a BusPipelineInfo containing a Pipeline ID, GameObject ID,
                 Object GUID and a dictionary containing additional information
                 (requested via `return_options`). When accessing the values in the dictionary, use the
                 EBusOptions enum as the keys. If this function call fails, an empty tuple is returned.
        """
        args = {"time": time}
        
        if bus_pipeline_id is not None:
            args["busPipelineId"] = bus_pipeline_id
        
        returns = list(returns) if returns is not None else list[EBusOptions]()
        returns.extend([EBusOptions.PIPELINE_ID, EBusOptions.GAME_OBJECT_ID, EBusOptions.OBJECT_GUID])
        options = {"return": list(dict.fromkeys(returns))}  # Remove duplicates
        
        results = self._client.call("ak.wwise.core.profiler.getBusses", args, options=options)
        results = results.get("return")
        
        if results is None:
            return tuple()
        
        objects = list[BusPipelineInfo]()
        
        for result in results:
            info = BusPipelineInfo(result[EBusOptions.PIPELINE_ID],
                                   result[EBusOptions.GAME_OBJECT_ID],
                                   result[EBusOptions.OBJECT_GUID],
                                   {k: v for k, v in result.items()
                                    if k not in returns})
            objects.append(info)
        
        return tuple(objects)
    
    def get_cpu_usage(self, time: ETimeCursor | int) -> tuple[CPUStatisticsInfo, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_profiler_getcpuusage.html \n
        Retrieves CPU usage statistics at a specific profiler capture time. This data can also be found
        in the Advanced Profiler, under the CPU tab. To ensure the CPU data is received,
        refer to `ak.wwise.core.profiler.enable_profiler_data`. The returned data includes "Inclusive" and
        "Exclusive" values, where "Inclusive" refers to the time spent in the element plus the time spent
        in any called elements, and "Exclusive" values pertain to execution only within the element
        itself.
        :param time: Time in milliseconds to query for cpu data, or a Time Cursor from which to acquire the time.
                     This parameter can have 2 possible values: int or ETimeCursor. The int is the time to query.
                     The ETimeCursor can have two values: user or capture. The User Time Cursor is the one that can
                     be manipulated by the user, while the Capture Time Cursor represents the latest time of the current
                     capture.
        :return: A tuple of CPUStatisticsInfo instances containing information about the amount of CPU percentage used
                 by each element.
        """
        args = {"time": time}
        
        results: dict[str, list[dict[str, int | float | str]]] = self._client.call("ak.wwise.core.profiler.getCpuUsage",
                                                                                   args)
        
        if results is None:
            return tuple[CPUStatisticsInfo, ...]()
        
        results: list[dict[str, int | float | str]] = results.get("return")  # Simplifying data structure.
        
        if not results:  # Empty.
            return tuple[CPUStatisticsInfo, ...]()
        
        stats = list[CPUStatisticsInfo]()
        
        for result in results:
            element_name = result.get("elementName", "")
            class_id = result.get("id", -1)
            instances = result.get("instances", -1)
            etype = result.get("type", "")
            percent_inclusive = result.get("percentInclusive", -1.0)
            percent_exclusive = result.get("percentExclusive", -1.0)
            ms_inclusive = result.get("millisecondsInclusive", -1.0)
            ms_exclusive = result.get("millisecondsExclusive", -1.0)
            stats.append(CPUStatisticsInfo(element_name, class_id, instances, etype, percent_inclusive,
                                           percent_exclusive, ms_inclusive, ms_exclusive))
        
        return tuple(stats)
    
    def get_cursor_time(self, time: ETimeCursor) -> int:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_profiler_getcursortime.html \n
        Returns the current time of the specified profiler cursor, in milliseconds.
        :param time: Time Cursor from which to acquire the time.
        :return: The current position of the specified Time Cursor, in ms. If function fails, it returns -1.
        """
        args = {"time": time}
        
        result = self._client.call("ak.wwise.core.profiler.getCursorTime", args)
        result = result.get("return")
        
        if result is None or result < 0:
            return -1
        
        return result
    
    def get_game_objects(self, time: ETimeCursor | int) -> tuple[GameObjectRegistrationData, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_profiler_getgameobjects.html \n
        Retrieves the game objects at a specific profiler capture time.
        :param time: The time in milliseconds to query for game objects. This parameter can have 2 possible values: int
                     or ETimeCursor. The int is the time to query. The ETimeCursor can have two values: user or capture.
                     The User Time Cursor is the one that can be manipulated by the user, while the Capture Time Cursor
                     represents the latest time of the current capture.
        :return: A tuple of objects containing game object registration data. When accessing the values in the
                 dictionary, use the EGameObjectRegistrationDataMembers enum as the keys. If this function call fails,
                 an empty tuple is returned.
        """
        args = {"time": time}
        
        results = self._client.call("ak.wwise.core.profiler.getGameObjects", args)
        
        if results is None:
            return tuple[GameObjectRegistrationData, ...]()
        
        results = results.get("return")
        
        if not results:  # Empty.
            return tuple[GameObjectRegistrationData, ...]()
        
        objects = list[GameObjectRegistrationData]()
        
        for result in results:
            ak_id = GameObjectID(result.get("id", GameObjectID.get_null()))
            name = result.get("name", "")
            register_time = result.get("registerTime", -1)
            unregister_time = result.get("unregisterTime", -1)
            objects.append(GameObjectRegistrationData(ak_id, name, register_time, unregister_time))
        
        return tuple(objects)
    
    def get_loaded_media(self, time: ETimeCursor | int) -> tuple[LoadedMediaInfo, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_profiler_getloadedmedia.html \n
        Retrieves the loaded media at a specific profiler capture time. This data can also be found in
        the Advanced Profiler, under the Loaded Media tab. To ensure the Loaded Media data is received,
        refer to `ak.wwise.core.profiler.enable_profiler_data`.
        :param time: Time in milliseconds to query for media, or a Time Cursor from which to acquire the time.
                     This parameter can have 2 possible values: int or ETimeCursor. The int is the time to query. The
                     ETimeCursor can have two values: user or capture. The User Time Cursor is the one that can be
                     manipulated by the user, while the Capture Time Cursor represents the latest time of the current
                     capture.
        :return: A tuple of loaded media information. When accessing the values in the dictionary, use the
                 ELoadedMediaMembers enum as the keys. If this function call fails, an empty tuple is returned.
        """
        args = {"time": time}
        
        results = self._client.call("ak.wwise.core.profiler.getLoadedMedia", args)
        results = results.get("return")
        
        if results is None:
            return tuple[LoadedMediaInfo, ...]()
        
        loaded_media = list[LoadedMediaInfo]()
        
        for result in results:
            media_id = ShortID(result.get("mediaId", ShortID.get_null()))
            file_name = Name(result.get("fileName", Name.get_null()))
            audio_format = result.get("format", "")
            size = result.get("size", -1)
            sound_bank = Name(result.get("soundBank", Name.get_null()))
            loaded_media.append(LoadedMediaInfo(media_id, file_name, audio_format, size, sound_bank))
        
        return tuple[LoadedMediaInfo, ...](loaded_media)
    
    def get_meters(self, time: ETimeCursor | int, returns: ListOrTuple[EReturnOptions] = None, platform: str = None,
                   language: str = None) -> tuple[WwiseObjectInfo, ...]:
        """
        https://www.audiokinetic.com/en/library/edge/?source=SDK&id=ak_wwise_core_profiler_getmeters.html \n
        Retrieves the Meter data for all registered buses, aux buses and devices. Only the master audio bus is
        registered by default. Use `ak.wwise.core.profiler.registerMeter` for other buses, before retrieval of the
        meter data.
        :param time: Time in milliseconds to query for media, or a Time Cursor from which to acquire the time.
                     This parameter can have 2 possible values: int or ETimeCursor. The int is the time to query. The
                     ETimeCursor can have two values: user or capture. The User Time Cursor is the one that can be
                     manipulated by the user, while the Capture Time Cursor represents the latest time of the current
                     capture.
        :param returns: The additional return options. By default, this function returns only the GUID and Name
                        of the selected objects. Duplicates are ignored.
        :param platform: If specified, this function will get information from the specified platform instead of the
                         current platform.
        :param language: If specified, this function will get information from the specified language instead of the
                         current language.
        :returns: For each selected object, a WwiseObjectInfo containing an object's GUID, name, path, type, and a
                  dictionary containing additional information (requested via `return_options`). When accessing the
                  values in the dictionary, use the EReturnOptions enum as the keys. If this function call fails, an
                  empty tuple is returned.
        """
        args = {"time": time}
        
        returns = list(dict.fromkeys(returns)) if returns is not None else list[EReturnOptions]()
        returns.extend(EReturnOptions.get_defaults())
        
        options = {"return": returns}
        
        if platform is not None:
            options["platform"] = platform
        if language is not None:
            options["language"] = language
        
        results = self._client.call("ak.wwise.core.profiler.getMeters", args, options=options)
        results = results.get("objects") if results is not None else None
        if results is None:
            return ()
        
        objects = list[WwiseObjectInfo]()
        
        for result in results:
            guid = GUID(result[EReturnOptions.GUID])
            name = Name(result[EReturnOptions.NAME])
            typename = EObjectType.from_type_name(result[EReturnOptions.TYPE])
            path = ProjectPath(result[EReturnOptions.PATH])
            other = {key: value for key, value in result.items() if key not in EReturnOptions.get_defaults()}
            objects.append(WwiseObjectInfo(guid, name, typename, path, other))
        
        return tuple(objects)
    
    def get_performance_monitor(self, time: ETimeCursor | int) -> tuple[PerformanceMonitorCounterInfo, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_profiler_getperformancemonitor.html \n
        Retrieves the Performance Monitor statistics at a specific profiler capture time. Refer to Wwise Authoring
        Performance Monitor Counter Identifiers for the available counters.
        :param time: Time in milliseconds to query for media, or a Time Cursor from which to acquire the time.
                     This parameter can have 2 possible values: int or ETimeCursor. The int is the time to query. The
                     ETimeCursor can have two values: user or capture. The User Time Cursor is the one that can be
                     manipulated by the user, while the Capture Time Cursor represents the latest time of the current
                     capture.
        :return: A tuple of performance monitor counter information. When accessing the values in the dictionary use the
                 EPerformanceMonitorMembers enum as the keys. If this function call fails, an empty tuple is returned.
        """
        args = {"time": time}
        
        results = self._client.call("ak.wwise.core.profiler.getPerformanceMonitor", args)
        
        if results is None:
            return tuple[PerformanceMonitorCounterInfo, ...]()
        
        results = results.get("return")
        
        if not results:  # Empty.
            return tuple[PerformanceMonitorCounterInfo, ...]()
        
        entries = list[PerformanceMonitorCounterInfo]()
        
        for result in results:
            name: str = result.get("name", "")
            identifier: str = result.get("id", "")
            value: float = result.get("value", float("-inf"))
            entries.append(PerformanceMonitorCounterInfo(name, identifier, value))
        
        return tuple[PerformanceMonitorCounterInfo, ...](entries)
    
    def get_rtpcs(self, time: ETimeCursor | int) -> tuple[ActiveRTPCInfo, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_profiler_getrtpcs.html \n
        Retrieves active RTPCs at a specific profiler capture time.
        :param time: Time in milliseconds to query for RTPCs, or a Time Cursor from which to acquire the time.
                     This parameter can have 2 possible values: int or ETimeCursor. The int is the time to query. The
                     ETimeCursor can have two values: user or capture. The User Time Cursor is the one that can be
                     manipulated by the user, while the Capture Time Cursor represents the latest time of the current
                     capture.
        :return: A tuple of RTPCs associated with a playing voice. When accessing the values in the dictionary use the
                 EActiveRTPCMembers enum as the keys. If this function call fails, an empty tuple is returned.
        """
        args = {"time": time}
        
        results = self._client.call("ak.wwise.core.profiler.getRTPCs", args)
        
        if results is None:
            return tuple[ActiveRTPCInfo, ...]()
        
        results = results.get("return", ())
        
        if not results:  # Empty.
            return tuple[ActiveRTPCInfo, ...]()
        
        rtpcs = list[ActiveRTPCInfo]()
        
        for result in results:
            guid = GUID(result.get("id", GUID.get_null()))
            name = Name(result.get("name", Name.get_null()))
            obj_id = GameObjectID(result.get("gameObjectId", GameObjectID.get_null()))
            value = result.get("value", float("-inf"))
            rtpcs.append(ActiveRTPCInfo(guid, name, obj_id, value))
        
        return tuple[ActiveRTPCInfo, ...](rtpcs)
    
    def get_streamed_media(self, time: ETimeCursor | int) -> tuple[StreamObjectInfo, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_profiler_getstreamedmedia.html \n
        Retrieves the streaming media at a specific profiler capture time. This data can also be found in
        the Advanced Profiler, under the Streams tab. To ensure the Streams data is received,
        refer to `ak.wwise.core.profiler.enable_profiler_data`.
        :param time: Time in milliseconds to query for RTPCs, or a Time Cursor from which to acquire the time.
                     This parameter can have 2 possible values: int or ETimeCursor. The int is the time to query. The
                     ETimeCursor can have two values: user or capture. The User Time Cursor is the one that can be
                     manipulated by the user, while the Capture Time Cursor represents the latest time of the current
                     capture.
        :return: Array of StreamObjectInfo, containing all data about how each of the streams is managed by the Wwise
                 sound engine.
        """
        args = {"time": time}
        
        results = self._client.call("ak.wwise.core.profiler.getStreamedMedia", args)
        results = results.get("return")
        
        if results is None:
            return tuple()
        
        streams: list[StreamObjectInfo] = [StreamObjectInfo(
            device_name=result.get("deviceName"),
            stream_name=result.get("streamName"),
            file_size=result.get("fileSize"),
            file_position=result.get("filePosition"),
            priority=result.get("priority"),
            bandwidth_total=result.get("bandwidthTotal"),
            bandwidth_low_level=result.get("bandwidthLowLevel"),
            referenced_memory=result.get("referencedMemory"),
            estimated_throughput=result.get("estimatedThroughput"),
            active=result.get("active"),
            target_buffer_size=result.get("targetBufferSize"),
            buffer_status_buffered=result.get("bufferStatusBuffered"),
        ) for result in results]
        
        return tuple(streams)
    
    def get_voice_contributions(self, time: ETimeCursor | int, voice_pipeline_id: int,
                                bus_pipeline_ids: ListOrTuple[int] = None) -> VoiceContributionHierarchy | None:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_profiler_getvoicecontributions.html \n
        Retrieves all parameters affecting voice volume, highpass and lowpass for a voice path,
        resolved from pipeline IDs.
        :param time: Time in milliseconds to query for RTPCs, or a Time Cursor from which to acquire the time.
                     This parameter can have 2 possible values: int or ETimeCursor. The int is the time to query. The
                     ETimeCursor can have two values: user or capture. The User Time Cursor is the one that can be
                     manipulated by the user, while the Capture Time Cursor represents the latest time of the current
                     capture.
        :param voice_pipeline_id: The pipeline ID of the voice to get contributions from. Identifies a playing voice
                                  instance ID
        :param bus_pipeline_ids: The pipeline IDs of buses belonging to a common voice path. An empty array defaults
                                 to the dry path. Identifies a playing voice instance ID
        :return: The hierarchy of objects with parameters contributing to the voice, ordered from top-level parent to
                 the voice object. If no information was found, `None` is returned.
        """
        args = {"time": time, "voicePipelineID": voice_pipeline_id}
        
        if bus_pipeline_ids is not None:
            args["bussesPipelineID"] = bus_pipeline_ids
        
        results = self._client.call("ak.wwise.core.profiler.getVoiceContributions", args)
        
        if results is None:
            return None
        
        results = results.get("return")
        
        if results is None:
            return None
        
        hierarchy = VoiceContributionHierarchy(results.get("volume", 0.0),
                                               results.get("LPF", 0.0),
                                               results.get("HPF", 0.0))
        
        def get_children(parent: VoiceContributionHierarchy | VoiceInspectorContribution, children: list[dict]):
            for child in children:
                inspector = VoiceInspectorContribution(child["name"], child["volume"], child["LPF"], child["HPF"])
                for param in child["parameters"]:
                    parameter = VoiceContributionParameter(param["propertyType"], param["reason"], param["driver"],
                                                           param["driverValue"], param["value"])
                    inspector.parameters.append(parameter)
                match parent:
                    case VoiceContributionHierarchy():
                        parent.objects.append(inspector)
                    case VoiceInspectorContribution():
                        parent.children.append(inspector)
                if len(child.get("children", ())) > 0:
                    get_children(inspector, child["children"])  # Recursive call
        
        get_children(hierarchy, results.get("objects", list[dict]()))
        
        if results is not None:
            return hierarchy
    
    def get_voices(self, time: ETimeCursor | int, voice_pipeline_id: int = None,
                   returns: ListOrTuple[EVoicePipelineReturnOptions] = None) -> tuple[PlayingVoiceProperties, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_profiler_getvoices.html \n
        Retrieves the voices at a specific profiler capture time.
        :param time: Time in milliseconds to query for RTPCs, or a Time Cursor from which to acquire the time.
                     This parameter can have 2 possible values: int or ETimeCursor. The int is the time to query. The
                     ETimeCursor can have two values: user or capture. The User Time Cursor is the one that can be
                     manipulated by the user, while the Capture Time Cursor represents the latest time of the current
                     capture.
        :param voice_pipeline_id: The pipeline ID of the voice to get contributions from. Identifies a playing voice
                                  instance ID
        :param returns: The additional return options. By default, this function returns only the Pipeline ID,
                        Game Object ID, and Object GUID.
        :return: An array of playing voices and their properties. Specify additional return options to extract more
                 data.
        """
        args = {"time": time}
        returns = list(returns) if returns is not None else list[EVoicePipelineReturnOptions]()
        returns.extend((EVoicePipelineReturnOptions.PIPELINE_ID,
                        EVoicePipelineReturnOptions.GAME_OBJECT_ID,
                        EVoicePipelineReturnOptions.OBJECT_GUID,))
        options = {"return": list(dict.fromkeys(returns))}  # Remove duplicates
        
        if voice_pipeline_id is not None:
            args["voicePipelineID"] = voice_pipeline_id
        
        results = self._client.call("ak.wwise.core.profiler.getVoices", args, options=options)
        results = results.get("return") if results is not None else None
        
        if results is None:
            return ()
        
        voices = list[PlayingVoiceProperties]()
        
        for voice in results:
            pipeline_id = int(voice[EVoicePipelineReturnOptions.PIPELINE_ID])
            game_object_id = GameObjectID(voice[EVoicePipelineReturnOptions.GAME_OBJECT_ID])
            object_guid = GUID(voice[EVoicePipelineReturnOptions.OBJECT_GUID])
            other = {key: value for key, value in voice.items() if key not in returns}
            voices.append(PlayingVoiceProperties(pipeline_id, game_object_id, object_guid, other))
        
        return tuple(voices)
    
    def register_meter(self, object_to_register: GUID | ProjectPath | str) -> bool:
        """
        https://www.audiokinetic.com/en/library/edge/?source=SDK&id=ak_wwise_core_profiler_registermeter.html \n
        Registers a bus, an aux bus or device to receive meter data. Only the master audio bus is registered by default.
        Use `ak.wwise.core.profiler.getMeters` to retrieve the meter data after registering. Every call to
        `ak.wwise.core.profiler.registerMeter` must have a matching call to `ak.wwise.core.profiler.unregisterMeter`.
        :param object_to_register: The ID (GUID), name, or path of the object to receive meter data. This object must
                                   be a bus, an aux bus or a device. The name of the object qualified by its type or
                                   Short ID in the form of type:name or Global:shortId. Only object types that have
                                   globally-unique names or Short Ids are supported. Ex: Event:Play_Sound_01
        :return: Returns a boolean that indicates if the call was successful.
        """
        result = self._client.call("ak.wwise.core.profiler.registerMeter", {"object": object_to_register})
        return True if result is not None else False
    
    def save_capture(self, file_path: SystemPath) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_profiler_savecapture.html \n
        Saves profiler as a .prof file according to the given file path.
        :param file_path: The file path to save the profiler to. Make sure to include .prof file extension.
                          E.g. C:\\MyProject\\capture.prof
        :return: True if the profiler was saved, False otherwise.
        """
        if file_path is None:
            return False
        
        file_path.parent.mkdir(parents=True, exist_ok=True)
        
        args = {"file": file_path}
        
        return self._client.call("ak.wwise.core.profiler.saveCapture", args) is not None
    
    def start_capture(self) -> int:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_profiler_startcapture.html \n
        Starts the profiler capture and returns the time at the beginning of the capture, in milliseconds.
        :return: The time at the beginning of the capture, in milliseconds. If method fails to call, it will return -1.
        """
        result = self._client.call("ak.wwise.core.profiler.startCapture", {})
        
        if result is not None:
            return result.get("return")
        else:
            return -1
    
    def stop_capture(self) -> int:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_core_profiler_stopcapture.html \n
        Stops the profiler capture and returns the time at the end of the capture, in milliseconds.
        :return: The time at the end of the capture, in milliseconds. If method fails to call, it will return -1.
        """
        result = self._client.call("ak.wwise.core.profiler.stopCapture", {})
        return result.get("return", -1) if result is not None else -1
    
    def unregister_meter(self, object_to_unregister: GUID | ProjectPath | str):
        """
        https://www.audiokinetic.com/en/library/edge/?source=SDK&id=ak_wwise_core_profiler_unregistermeter.html \n
        Unregisters a bus or device that was registered with `ak.wwise.core.profiler.registerMeter`.
        :param object_to_unregister: The ID (GUID), name, or path of the object to receive meter data. This object must
                                     be a bus, an aux bus or a device. The name of the object qualified by its type or
                                     Short ID in the form of type:name or Global:shortId. Only object types that have
                                     globally-unique names or Short Ids are supported. Ex: Event:Play_Sound_01
        :return: Returns a boolean that indicates if the call was successful.
        """
        return self._client.call("ak.wwise.core.profiler.unregisterMeter",
                                 {"object": object_to_unregister}) is not None



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.enums import EColour, EFadeShape, EMarkerInputMode, ESpeakerBitMask
from pywwise.objects.abc import WwiseObject
from pywwise.objects.types.conversion import Conversion
from pywwise.objects.types.marker import Marker


class AudioSource(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_audiofilesource.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.AUDIO_SOURCE`.
    """
    channel_config_override = WwiseProperty[ESpeakerBitMask]("ChannelConfigOverride", ESpeakerBitMask)
    colour = WwiseProperty[EColour]("Color", EColour)
    conversion = WwiseProperty[Conversion]("Conversion", Conversion)
    crossfade_duration = WwiseProperty[float]("CrossfadeDuration", float)
    crossfade_shape = WwiseProperty[EFadeShape]("CrossfadeShape", EFadeShape)
    fade_in_duration = WwiseProperty[float]("FadeInDuration", float)
    fade_in_shape = WwiseProperty[EFadeShape]("FadeInShape", EFadeShape)
    fade_out_duration = WwiseProperty[float]("FadeOutDuration", float)
    fade_out_shape = WwiseProperty[EFadeShape]("FadeOutShape", EFadeShape)
    hdr_envelope = WwiseProperty[float]("HdrEnvelope", float)
    loop_begin = WwiseProperty[float]("LoopBegin", float)
    loop_end = WwiseProperty[float]("LoopEnd", float)
    marker_detection_sensitivity = WwiseProperty[float]("MarkerDetectionSensitivity", float)
    marker_input_mode = WwiseProperty[EMarkerInputMode]("MarkerInputMode", EMarkerInputMode)
    markers = WwiseProperty[list[Marker]]("Markers", list[Marker])
    override_colour = WwiseProperty[bool]("OverrideColor", bool)
    override_conversion = WwiseProperty[bool]("OverrideConversion", bool)
    override_wav_loop = WwiseProperty[bool]("OverrideWavLoop", bool)
    trim_begin = WwiseProperty[float]("TrimBegin", float)
    trim_end = WwiseProperty[float]("TrimEnd", float)
    volume_offset = WwiseProperty[float]("VolumeOffset", float)



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.enums import (EReflectAlgorithm, EReflectChannelConfig, EReflectDecorrelationStrengthSource,
                           EReflectFilterType, EReflectThresholdMode)
from pywwise.objects.abc import WwiseObject


class WwiseReflect(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_effect_wwise_reflect.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.WWISE_REFLECT`.
    """
    base_texture_frequency = WwiseProperty[float]("BaseTextureFrequency", float)
    centre_percentage = WwiseProperty[float]("CenterPerc", float)
    curve_usage_mask = WwiseProperty[int]("CurveUsageMask", int)
    decorrelation_algorithm = WwiseProperty[EReflectAlgorithm]("DecorrAlgorithmSelect", EReflectAlgorithm)
    decorrelation_filter_max_reflection_order = WwiseProperty[int]("DecorrFilterMaxReflectionOrder", int)
    decorrelation_hardware_acceleration = WwiseProperty[bool]("DecorrHardwareAcceleration", bool)
    decorrelation_strength = WwiseProperty[float]("DecorrStrength", float)
    decorrelation_strength_source = WwiseProperty[EReflectDecorrelationStrengthSource](
        "DecorrStrengthSource", EReflectDecorrelationStrengthSource)
    decorrelation_window_width = WwiseProperty[int]("DecorrWindowWidth", int)
    diffraction_warping = WwiseProperty[float]("DiffractionWarping", float)
    distance_threshold = WwiseProperty[float]("DistanceThreshold", float)
    distance_warping = WwiseProperty[float]("DistanceWarping", float)
    dry = WwiseProperty[float]("Dry", float)
    fade_time = WwiseProperty[float]("FadeTime", float)
    fusing_time = WwiseProperty[float]("FusingTime", float)
    material_filtering_select = WwiseProperty[EReflectAlgorithm]("MaterialFilteringSelect", EReflectAlgorithm)
    max_distance = WwiseProperty[float]("MaxDistance", float)
    max_reflections = WwiseProperty[float]("MaxReflections", float)
    output_channel_config = WwiseProperty[EReflectChannelConfig]("OutputChannelConfig", EReflectChannelConfig)
    param_filter_cutoff = WwiseProperty[float]("ParamFilterCutoff", float)
    param_filter_type = WwiseProperty[EReflectFilterType]("ParamFilterType", EReflectFilterType)
    pitch_threshold = WwiseProperty[float]("PitchThreshold", float)
    speed_of_sound = WwiseProperty[float]("SpeedOfSound", float)
    stereo_decorrelation = WwiseProperty[bool]("StereoDecorrelation", bool)
    threshold_mode = WwiseProperty[EReflectThresholdMode]("ThresholdMode", EReflectThresholdMode)
    wet = WwiseProperty[float]("Wet", float)



# Copyright 2024 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from typing import Self as _Self, TypeAlias as _TypeAlias

from waapi import CallbackExecutor, SequentialThreadExecutor, WaapiClient as _WaapiClient

from pywwise.aliases import ListOrTuple
from pywwise.structs import WwiseObjectWatch
from pywwise.waapi.ak.soundengine import SoundEngine as _SoundEngine
from pywwise.waapi.ak.wwise import Wwise as _Wwise


class Ak:
    """ak"""
    
    def __init__(self, url: str = "ws://127.0.0.1:8080/waapi", allow_exception: bool = False,
                 callback_executor: CallbackExecutor = SequentialThreadExecutor,
                 is_debug_build: bool = False, is_console_instance: bool = False,
                 watch_list: ListOrTuple[WwiseObjectWatch] = ()):
        """
        Constructor.
        :param url: URL of the Wwise Authoring API WAMP server, defaults to `ws://127.0.0.1:8080/waapi`.
        :param allow_exception: Allow errors on call and subscribe to throw an exception. Default is False.
        :param callback_executor: Executor strategy for event callbacks.
        :param is_debug_build: Should be set to true if the instance of Wwise is a debug build and debug-only.
        :param is_console_instance: Should be set to true if the instance of Wwise is running in a console window.
                                    functions/topics are required.
        :param watch_list: A tuple of `WwiseObjectWatch` instances. This will be used to set up the
                           `ak.wwise.core.object.property_changed` event.
        """
        self._client = _WaapiClient(url, allow_exception, callback_executor)
        self.soundengine = _SoundEngine(self._client)
        self.wwise = _Wwise(self._client, is_debug_build, is_console_instance, watch_list)
    
    def __enter__(self) -> _Self:
        """
        Enter the context (re: `with` statement).
        :return: This instance of the `Ak` class.
        """
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        """
        Exit the context (re: `with` statement).
        :param exc_type: The exception type, if any.
        :param exc_value: The exception value, if any.
        :param traceback: The traceback, if any exception(s) were raised.
        :return: Whether an exception was raised.
        """
        self._client.disconnect()
        return bool(exc_type)
    
    def is_connected(self):
        """
        Check if this instance is connected to Wwise.
        :return: Whether this instance is connected to Wwise.
        """
        return self._client.is_connected()
    
    def disconnect(self):
        """Disconnect from Wwise."""
        self._client.disconnect()


WwiseConnection: _TypeAlias = Ak  # This cannot exist in aliases.py due to a circular import issue.
"""Represents a connection to Wwise."""



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.enums import ERecorderAmbisonicsChannelOrdering, ERecorderFormat
from pywwise.objects.abc import WwiseObject


class WwiseRecorder(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_effect_wwise_recorder.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.WWISE_RECORDER`.
    """
    ambisonics_channel_ordering = WwiseProperty[ERecorderAmbisonicsChannelOrdering](
        "AmbisonicsChannelOrdering", ERecorderAmbisonicsChannelOrdering)
    apply_downstream_volume = WwiseProperty[bool]("ApplyDownstreamVolume", bool)
    authoring_filename = WwiseProperty[str]("AuthoringFilename", str)
    centre = WwiseProperty[float]("Center", float)
    downmix_to_stereo = WwiseProperty[bool]("DownmixToStereo", bool)
    format = WwiseProperty[ERecorderFormat]("Format", ERecorderFormat)
    game_filename = WwiseProperty[str]("GameFilename", str)
    lfe = WwiseProperty[float]("LFE", float)
    rear = WwiseProperty[float]("Rear", float)
    surround = WwiseProperty[float]("Surround", float)



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.objects.abc import WwiseObject


class WwiseGain(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_effect_wwise_gain.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.WwiseGain`.
    """
    full_band_gain = WwiseProperty[float]("FullBandGain", float)
    lfe_gain = WwiseProperty[float]("LFEGain", float)



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.enums import EPitchShifterFilterType, EPitchShifterInput
from pywwise.objects.abc import WwiseObject


class WwisePitchShifter(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_effect_wwise_pitch_shifter.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.WWISE_PITCH_SHIFTER`.
    """
    delay_time = WwiseProperty[float]("DelayTime", float)
    dry_level = WwiseProperty[float]("DryLevel", float)
    filter_frequency = WwiseProperty[float]("FilterFrequency", float)
    filter_gain = WwiseProperty[float]("FilterGain", float)
    filter_q = WwiseProperty[float]("FilterQFactor", float)
    filter_type = WwiseProperty[EPitchShifterFilterType]("FilterType", EPitchShifterFilterType)
    input = WwiseProperty[EPitchShifterInput]("Input", EPitchShifterInput)
    pitch = WwiseProperty[float]("Pitch", float)
    process_lfe = WwiseProperty[bool]("ProcessLFE", bool)
    sync_dry = WwiseProperty[bool]("SyncDry", bool)
    wet_level = WwiseProperty[float]("WetLevel", float)



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.enums import EChannelConversionSettings, EColour, ESampleRate, ESampleRateConversionQuality
from pywwise.objects.abc import WwiseObject


class Conversion(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_conversion.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.CONVERSION`.
    """
    allow_channel_upmix = WwiseProperty[bool]("AllowChannelUpmix", bool)
    channels = WwiseProperty[EChannelConversionSettings]("Channels", EChannelConversionSettings)
    colour = WwiseProperty[EColour]("Color", EColour)
    lr_mix = WwiseProperty[float]("LRMix", float)
    max_sample_rate = WwiseProperty[ESampleRate]("MaxSampleRate", ESampleRate)
    min_sample_rate = WwiseProperty[ESampleRate]("MinSampleRate", ESampleRate)
    override_colour = WwiseProperty[bool]("OverrideColor", bool)
    remove_dc_offset = WwiseProperty[bool]("RemoveDCOffset", bool)
    sr_conversion_quality = WwiseProperty[ESampleRateConversionQuality](
        "SRConversionQuality", ESampleRateConversionQuality)
    sample_rate = WwiseProperty[ESampleRate]("SampleRate", ESampleRate)
    use_dither = WwiseProperty[bool]("UseDither", bool)
    use_filename_marker = WwiseProperty[bool]("UseFilenameMarker", bool)



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from importlib import import_module as _import_module


class LazyModule:
    """A proxy for lazy-importing modules."""
    
    def __init__(self, module_name: str):
        """
        Stores a module name, which will be lazy-initialized (deferred).
        :param module_name: The name of the module.
        """
        self.module_name = module_name
        self._module = None
    
    def __getattr__(self, name):
        """
        During the first access,
        :param name:
        :return:
        """
        if self._module is None:
            self._module = _import_module(self.module_name)
        return getattr(self._module, name)



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.enums import EColour, EMatchMode
from pywwise.objects.abc import WwiseObject
from pywwise.objects.types.multi_switch_entry import MultiSwitchEntry
from pywwise.objects.types.state_group import StateGroup
from pywwise.objects.types.switch_group import SwitchGroup


class DialogueEvent(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_dialogueevent.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.DIALOGUE_EVENT`.
    """
    arguments = WwiseProperty[tuple[SwitchGroup | StateGroup, ...]]("Arguments", tuple)
    colour = WwiseProperty[EColour]("Color", EColour)
    entries = WwiseProperty[tuple[MultiSwitchEntry, ...]]("Entries", tuple)
    mode = WwiseProperty[EMatchMode]("Mode", EMatchMode)
    override_colour = WwiseProperty[bool]("OverrideColor", bool)
    probability = WwiseProperty[int]("Probability", int)



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.objects.abc import WwiseObject


class Impacter(WwiseObject):  # TODO: Implement enums for this class once the plugin is not in beta anymore.
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_source_impacter.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.SOURCE_PLUGIN`. Note: this plugin is in BETA and subject to change. Some
    properties might also not work as expected or even be displayed in the Wwise UI.
    """
    excitation_mask = WwiseProperty[int]("ExcitationMask", int)
    excitation_selection_mode = WwiseProperty[int]("ExcitationSelectionMode", int)
    fm_depth = WwiseProperty[float]("FMDepth", float)
    fm_depth_random = WwiseProperty[float]("FMDepthRandom", float)
    file_excitation = WwiseProperty[int]("FileExcitation", int)
    file_model = WwiseProperty[int]("FileModel", int)
    impact_position = WwiseProperty[float]("ImpactPosition", float)
    impact_position_random = WwiseProperty[float]("ImpactPositionRandom", float)
    mass = WwiseProperty[float]("Mass", float)
    mass_random = WwiseProperty[float]("MassRandom", float)
    min_duration = WwiseProperty[float]("MinDuration", float)
    model_mask = WwiseProperty[int]("ModelMask", int)
    model_selection_mode = WwiseProperty[int]("ModelSelectionMode", int)
    num_files = WwiseProperty[int]("NumFiles", int)
    output_level = WwiseProperty[float]("OutputLevel", float)
    velocity = WwiseProperty[float]("Velocity", float)
    velocity_random = WwiseProperty[float]("VelocityRandom", float)



# Copyright 2024 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from simplevent import RefEvent as _RefEvent
from waapi import WaapiClient as _WaapiClient

from pywwise.decorators import callback
from pywwise.enums import ECommand, EObjectType, EReturnOptions
from pywwise.primitives import GUID, Name, ProjectPath, ShortID
from pywwise.statics import EnumStatics
from pywwise.structs import CommandInfo, PlatformInfo, WwiseObjectInfo


class Commands:
    """ak.wwise.ui.commands"""
    
    def __init__(self, client: _WaapiClient):
        """
        Constructor.
        :param client: The WAAPI client to use.
        """
        self._client = client
        
        self.executed = _RefEvent(ECommand, tuple[WwiseObjectInfo, ...], tuple[str, ...])
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_ui_commands_executed.html
        \nSent when a command is executed. The objects for which the command is executed are sent in the publication.
        \n**Event Data**:
        \n- The command that was executed.
        \n- A tuple of objects (WwiseObjectInfo instances) for which the command was executed. May be empty.
        \n- A tuple of platforms (GUID or name, as a string) for which the command was executed.
        """
        
        executed_options = {"return": [EReturnOptions.GUID.value, EReturnOptions.NAME.value,
                                       EReturnOptions.TYPE.value, EReturnOptions.PATH.value]}
        self._executed = self._client.subscribe("ak.wwise.ui.commands.executed", self._on_executed,
                                                executed_options)
    
    @callback
    def _on_executed(self, event: _RefEvent, **kwargs):
        """
        Callback function for the `executed` event.
        :param event: The event to broadcast.
        :param kwargs: The event data.
        """
        command = EnumStatics.from_value(ECommand, kwargs["command"])
        objs = tuple([WwiseObjectInfo.from_dict(obj) for obj in kwargs["objects"]])
        platforms = tuple(kwargs["platforms"])
        event(command, objs, platforms)
    
    def get_commands(self) -> tuple[str, ...]:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_ui_commands_getcommands.html \n
        Gets the list of commands.
        :return: A tuple containing the available commands. If empty, the call failed.
        """
        commands = self._client.call("ak.wwise.ui.commands.getCommands").get("commands")
        return tuple(commands) if commands is not None else ()
    
    def execute(self, command: ECommand,
                objects: tuple[WwiseObjectInfo | GUID | ProjectPath | ShortID | tuple[EObjectType, Name]] = None,
                platforms: set[PlatformInfo | Name | GUID] = None, value: str | float | bool = None) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_ui_commands_execute.html \n
        Executes a command. Some commands can take a list of objects as parameters. Refer to Wwise
        Authoring Command Identifiers for the available commands.
        :param command: The ID of the command to execute. Refer to Wwise Authoring Command Identifiers for the lists of
                        commands.
        :param objects: For commands that take objects as arguments. This set should contain the GUID or project path
                        of the Wwise objects being passed to the command to be executed. Name and ShortID are both
                        supported, but are NOT recommended, as they will only work with "global" types (e.g. State
                        Groups, Events, etc.). To use a Name, you will need to pass a tuple containing the object type
                        and the actual name; for ShortID, only "global" types (e.g. State Groups) will work.
        :param platforms: An array of platform. Each platform is a GUID or a Name. Some commands can take
                          platforms as arguments. Refer to the commands for more information.
        :param value: A value to pass to the command. Some commands can take a value as an argument. **Can be Null,
                      String, Float, or Bool**. Refer to the commands for more information.
        :return: Whether the call succeeded.
        """
        if command is None:
            return False
        
        args = {"command": command, "objects": list(), "platforms": list(), "value": None}
        
        if objects is not None:
            for wwise_object in objects:
                match wwise_object:
                    case WwiseObjectInfo():
                        args["objects"].append(wwise_object.guid)
                    case GUID() | ProjectPath() | Name():
                        args["objects"].append(wwise_object)
                    case ShortID():  # "Global:0123456789"
                        args["objects"].append(f"Global:{wwise_object}")
                    case tuple():  # "Type:Name"
                        args["objects"].append(f"{wwise_object[0].value}:{wwise_object[1]}")
                    case _:
                        args["objects"].append(str(wwise_object))
        if platforms is not None:
            for platform in platforms:
                match platform:
                    case GUID() | Name():
                        args["platforms"].append(platform)
                    case PlatformInfo():
                        args["platforms"].append(platform.guid)
        if value is not None:
            args["value"] = value
        
        return self._client.call("ak.wwise.ui.commands.execute", args) is not None
    
    def register(self, commands: set[CommandInfo]) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_ui_commands_register.html \n
        Registers an array of add-on commands. Registered commands remain until the Wwise process is
        terminated. Refer to Defining Command Add-ons for more information about registering commands.
        Also refer to `ak.wwise.ui.commands.executed`.
        :param commands: An array of add-on commands. Data for the commands to be registered.
        :return: Whether the call succeeded.
        """
        if commands is None:
            return False
        
        args = {"commands": list()}
        
        for command in commands:
            args["commands"].append(command.dictionary)
        
        return self._client.call("ak.wwise.ui.commands.register", args) is not None
    
    def unregister(self, commands: set[CommandInfo | str]) -> bool:
        """
        https://www.audiokinetic.com/library/edge/?source=SDK&id=ak_wwise_ui_commands_unregister.html \n
        Unregisters an array of add-on UI commands.
        :param commands: A set of add-on commands. Can either be a string containing the add-on command ID or a full
                         Add-on Command Info instance.
        :return: Whether the call succeeded.
        """
        if commands is None:
            return False
        
        args = {"commands": list()}
        
        for command in commands:
            if isinstance(command, CommandInfo):
                args["commands"].append(command.id)
            else:
                args["commands"].append(str(command))
        
        return self._client.call("ak.wwise.ui.commands.unregister", args) is not None



# Copyright 2025 Matheus Vilano
# SPDX-License-Identifier: Apache-2.0

from pywwise.descriptors import WwiseProperty
from pywwise.enums import EFadeType, EMusicFadeShape
from pywwise.objects.abc import WwiseObject


class MusicFade(WwiseObject):
    """
    https://www.audiokinetic.com/en/library/edge/?source=SDK&id=wwiseobject_musicfade.html \n
    A class serving as an interface for getting/setting properties on Wwise objects. This type specifically targets
    the class represented by `EObjectType.MUSIC_FADE`.
    """
    fade_curve = WwiseProperty[EMusicFadeShape]("FadeCurve", EMusicFadeShape)
    fade_offset = WwiseProperty[float]("FadeOffset", float)
    fade_time = WwiseProperty[float]("FadeTime", float)
    fade_type = WwiseProperty[EFadeType]("FadeType", EFadeType)
